from infrahub.core.branch.needs_rebase_status import check_need_rebase_status

ALLOWED_MUTATIONS_ON_NEED_REBASE_BRANCH = ["BranchRebase", "BranchDelete", "BranchCreate", "ProposedChangeCreate"]


def raise_on_mutation_on_branch_needing_rebase(next, root, info, **kwargs):  # type: ignore  # noqa
    if info.operation.operation.value == "mutation":
        mutation_name = info.operation.selection_set.selections[0].name.value
        if mutation_name not in ALLOWED_MUTATIONS_ON_NEED_REBASE_BRANCH:
            check_need_rebase_status(branch=info.context.branch)

    return next(root, info, **kwargs)



from infrahub.log import get_logger
from infrahub.message_bus import messages
from infrahub.tasks.registry import refresh_branches
from infrahub.worker import WORKER_IDENTITY
from infrahub.workers.dependencies import get_component, get_database


async def branches(message: messages.RefreshRegistryBranches) -> None:
    if message.meta and message.meta.initiator_id == WORKER_IDENTITY:
        get_logger().info("Ignoring refresh registry refresh request originating from self", worker=WORKER_IDENTITY)
        return

    database = await get_database()
    async with database.start_session(read_only=False) as db:
        await refresh_branches(db=db)

    component = await get_component()
    await component.refresh_schema_hash()


async def rebased_branch(message: messages.RefreshRegistryRebasedBranch) -> None:
    if message.meta and message.meta.initiator_id == WORKER_IDENTITY:
        get_logger().info(
            "Ignoring refresh registry refreshed branch for request originating from self", worker=WORKER_IDENTITY
        )
        return

    database = await get_database()

    async with database.start_session(read_only=True) as db:
        await refresh_branches(db=db)

    component = await get_component()
    await component.refresh_schema_hash()



from infrahub.core.constraint.node.runner import NodeConstraintRunner
from infrahub.dependencies.interface import DependencyBuilder, DependencyBuilderContext

from ..node.grouped_uniqueness import NodeGroupedUniquenessConstraintDependency
from ..relationship_manager.count import RelationshipCountConstraintDependency
from ..relationship_manager.peer_kind import RelationshipPeerKindConstraintDependency
from ..relationship_manager.peer_parent import RelationshipPeerParentConstraintDependency
from ..relationship_manager.peer_relatives import RelationshipPeerRelativesConstraintDependency
from ..relationship_manager.profiles_kind import RelationshipProfilesKindConstraintDependency


class NodeConstraintRunnerDependency(DependencyBuilder[NodeConstraintRunner]):
    @classmethod
    def build(cls, context: DependencyBuilderContext) -> NodeConstraintRunner:
        return NodeConstraintRunner(
            db=context.db,
            branch=context.branch,
            uniqueness_constraint=NodeGroupedUniquenessConstraintDependency.build(context=context),
            relationship_manager_constraints=[
                RelationshipPeerKindConstraintDependency.build(context=context),
                RelationshipCountConstraintDependency.build(context=context),
                RelationshipProfilesKindConstraintDependency.build(context=context),
                RelationshipPeerParentConstraintDependency.build(context=context),
                RelationshipPeerRelativesConstraintDependency.build(context=context),
            ],
        )



from __future__ import annotations

import re
from typing import TYPE_CHECKING, Any, Optional, Self, Union

from pydantic import Field, field_validator

from infrahub.core.branch.enums import BranchStatus
from infrahub.core.constants import GLOBAL_BRANCH_NAME
from infrahub.core.graph import GRAPH_VERSION
from infrahub.core.models import SchemaBranchHash  # noqa: TC001
from infrahub.core.node.standard import StandardNode
from infrahub.core.query import QueryType
from infrahub.core.query.branch import (
    DeleteBranchRelationshipsQuery,
    GetAllBranchInternalRelationshipQuery,
    RebaseBranchDeleteRelationshipQuery,
    RebaseBranchUpdateRelationshipQuery,
)
from infrahub.core.registry import registry
from infrahub.core.timestamp import Timestamp
from infrahub.exceptions import BranchNotFoundError, InitializationError, ValidationError

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase


class Branch(StandardNode):
    name: str = Field(
        max_length=250, min_length=3, description="Name of the branch (git ref standard)", validate_default=True
    )
    status: BranchStatus = BranchStatus.OPEN
    description: str = ""
    origin_branch: str = "main"
    branched_from: Optional[str] = Field(default=None, validate_default=True)
    hierarchy_level: int = 2
    created_at: Optional[str] = Field(default=None, validate_default=True)
    is_default: bool = False
    is_global: bool = False
    is_protected: bool = False
    sync_with_git: bool = Field(
        default=True,
        description="Indicate if the branch should be extended to Git and if Infrahub should merge the branch in Git as part of a proposed change",
    )
    is_isolated: bool = True
    schema_changed_at: Optional[str] = None
    schema_hash: Optional[SchemaBranchHash] = None
    graph_version: int | None = None

    _exclude_attrs: list[str] = ["id", "uuid", "owner"]

    @field_validator("name", mode="before")
    @classmethod
    def validate_branch_name(cls, value: str) -> str:
        checks = [
            (r".*/\.", "/."),
            (r"\.\.", ".."),
            (r"^/", "starts with /"),
            (r"//", "//"),
            (r"@{", "@{"),
            (r"\\", "backslash (\\)"),
            (r"[\000-\037\177 ~^:?*[]", "disallowed ASCII characters/patterns"),
            (r"\.lock$", "ends with .lock"),
            (r"/$", "ends with /"),
            (r"\.$", "ends with ."),
        ]

        offending_patterns = []

        for pattern, description in checks:
            if re.search(pattern, value):
                offending_patterns.append(description)

        if value == GLOBAL_BRANCH_NAME:
            return value  # this is the only allowed exception

        if offending_patterns:
            error_text = ", ".join(offending_patterns)
            raise ValidationError(f"Branch name contains invalid patterns or characters: {error_text}")

        return value

    @field_validator("branched_from", mode="before")
    @classmethod
    def set_branched_from(cls, value: str) -> str:
        return Timestamp(value).to_string()

    def get_branched_from(self) -> str:
        if not self.branched_from:
            raise RuntimeError(f"branched_from not set for branch {self.name}")
        return self.branched_from

    @field_validator("created_at", mode="before")
    @classmethod
    def set_created_at(cls, value: str) -> str:
        return Timestamp(value).to_string()

    def get_created_at(self) -> str:
        if not self.created_at:
            raise RuntimeError(f"created_at not set for branch {self.name}")
        return self.created_at

    @property
    def active_schema_hash(self) -> SchemaBranchHash:
        if self.schema_hash:
            return self.schema_hash

        raise InitializationError("The schema_hash has not been loaded for this branch")

    @property
    def has_schema_changes(self) -> bool:
        if not self.schema_hash:
            return False

        origin_branch = self.get_origin_branch()
        if not origin_branch or not origin_branch.schema_hash:
            return False

        if self.schema_hash.main != origin_branch.schema_hash.main:
            return True

        return False

    def update_schema_hash(self, at: Timestamp | str | None = None) -> bool:
        latest_schema = registry.schema.get_schema_branch(name=self.name)
        new_hash = latest_schema.get_hash_full()
        if self.schema_hash and new_hash.main == self.schema_hash.main:
            return False

        self.schema_changed_at = Timestamp(at).to_string()
        self.schema_hash = new_hash
        return True

    @classmethod
    async def get_by_name(cls, name: str, db: InfrahubDatabase, ignore_deleting: bool = True) -> Branch:
        query = """
        MATCH (n:Branch)
        WHERE n.name = $name
        AND NOT n.status IN $ignore_statuses
        RETURN n
        """

        params: dict[str, Any] = {"name": name}
        params["ignore_statuses"] = []
        if ignore_deleting:
            params["ignore_statuses"].append(BranchStatus.DELETING.value)

        results = await db.execute_query(query=query, params=params, name="branch_get_by_name", type=QueryType.READ)

        if len(results) == 0:
            raise BranchNotFoundError(identifier=name)

        return cls.from_db(results[0].values()[0])

    @classmethod
    async def get_list(
        cls,
        db: InfrahubDatabase,
        limit: int = 1000,
        ids: list[str] | None = None,
        name: str | None = None,
        **kwargs: dict[str, Any],
    ) -> list[Self]:
        branches = await super().get_list(db=db, limit=limit, ids=ids, name=name, **kwargs)
        branches = [branch for branch in branches if branch.status != BranchStatus.DELETING]

        return branches

    @classmethod
    def isinstance(cls, obj: Any) -> bool:
        return isinstance(obj, cls)

    def get_origin_branch(self) -> Optional[Branch]:
        """Return the branch Object of the origin_branch."""
        if not self.origin_branch or self.origin_branch == self.name:
            return None

        return registry.get_branch_from_registry(branch=self.origin_branch)

    def get_branches_in_scope(self) -> list[str]:
        """Return the list of all the branches that are constituing this branch.

        For now, either a branch is the default branch or it must inherit from it so we can only have 2 values at best
        But the idea is that it will change at some point in a future version.
        """
        default_branch = registry.default_branch
        if self.name == default_branch:
            return [self.name]

        return [default_branch, self.name]

    def get_branches_and_times_to_query(self, at: Optional[Union[Timestamp, str]] = None) -> dict[frozenset, str]:
        """Return all the names of the branches that are constituing this branch with the associated times excluding the global branch"""

        at = Timestamp(at)

        if self.is_default:
            return {frozenset([self.name]): at.to_string()}

        time_default_branch = at

        # If the branch is isolated, and if the time requested is after the creation of the branch
        if self.is_isolated and at > Timestamp(self.branched_from):
            time_default_branch = Timestamp(self.branched_from)

        return {
            frozenset([self.origin_branch]): time_default_branch.to_string(),
            frozenset([self.name]): at.to_string(),
        }

    def get_branches_and_times_to_query_global(
        self,
        at: Optional[Union[Timestamp, str]] = None,
        is_isolated: bool = True,
    ) -> dict[frozenset, str]:
        """Return all the names of the branches that are constituting this branch with the associated times."""

        at = Timestamp(at)

        if self.is_default:
            return {frozenset((GLOBAL_BRANCH_NAME, self.name)): at.to_string()}

        time_default_branch = at

        # If the branch is isolated, and if the time requested is after the creation of the branch
        if self.is_isolated and is_isolated and at > Timestamp(self.branched_from):
            time_default_branch = Timestamp(self.branched_from)

        return {
            frozenset((GLOBAL_BRANCH_NAME, self.origin_branch)): time_default_branch.to_string(),
            frozenset((GLOBAL_BRANCH_NAME, self.name)): at.to_string(),
        }

    def get_branches_and_times_for_range(
        self, start_time: Timestamp, end_time: Timestamp
    ) -> tuple[dict[str, str], dict[str, str]]:
        """Return the names of the branches that are constituing this branch with the start and end times."""

        start = {}
        end = {}

        time_branched_from = Timestamp(self.branched_from)
        time_created_at = Timestamp(self.created_at)

        # Ensure start time is not older than the creation of the branch (time_created_at)
        time_query_start = start_time
        if start_time < time_created_at:
            time_query_start = time_created_at

        start[self.name] = time_query_start.to_string()

        # START
        if not self.is_default and time_query_start <= time_branched_from:
            start[self.origin_branch] = time_branched_from.to_string()
        elif not self.is_default and time_query_start > time_branched_from:
            start[self.origin_branch] = time_query_start.to_string()

        # END
        end[self.name] = end_time.to_string()
        if not self.is_default:
            end[self.origin_branch] = end_time.to_string()

        return start, end

    async def create(self, db: InfrahubDatabase) -> bool:
        self.graph_version = GRAPH_VERSION
        return await super().create(db=db)

    async def delete(self, db: InfrahubDatabase) -> None:
        if self.is_default:
            raise ValidationError(f"Unable to delete {self.name} it is the default branch.")
        if self.is_global:
            raise ValidationError(f"Unable to delete {self.name} this is an internal branch.")

        self.status = BranchStatus.DELETING
        await self.save(db=db)

        query = await DeleteBranchRelationshipsQuery.init(db=db, branch_name=self.name)
        await query.execute(db=db)
        await super().delete(db=db)

    def get_query_filter_relationships(
        self, rel_labels: list, at: Optional[Union[Timestamp, str]] = None, include_outside_parentheses: bool = False
    ) -> tuple[list, dict]:
        """
        Generate a CYPHER Query filter based on a list of relationships to query a part of the graph at a specific time and on a specific branch.
        """

        filters = []
        params: dict[str, Any] = {}

        if not isinstance(rel_labels, list):
            raise TypeError(f"rel_labels must be a list, not a {type(rel_labels)}")

        at = Timestamp(at)
        branches_times = self.get_branches_and_times_to_query_global(at=at)

        for idx, (branch_name, time_to_query) in enumerate(branches_times.items()):
            params[f"branch{idx}"] = list(branch_name)
            params[f"time{idx}"] = time_to_query

        for rel in rel_labels:
            filters_per_rel = []
            for idx in range(len(branches_times)):
                filters_per_rel.append(
                    f"({rel}.branch IN $branch{idx} AND {rel}.from <= $time{idx} AND {rel}.to IS NULL)"
                )
                filters_per_rel.append(
                    f"({rel}.branch IN $branch{idx} AND {rel}.from <= $time{idx} AND {rel}.to >= $time{idx})"
                )

            if not include_outside_parentheses:
                filters.append("\n OR ".join(filters_per_rel))

            filters.append("(" + "\n OR ".join(filters_per_rel) + ")")

        return filters, params

    def get_query_filter_path(
        self,
        at: Optional[Union[Timestamp, str]] = None,
        is_isolated: bool = True,
        branch_agnostic: bool = False,
        variable_name: str = "r",
        params_prefix: str = "",
    ) -> tuple[str, dict]:
        """
        Generate a CYPHER Query filter based on a path to query a part of the graph at a specific time and on a specific branch.

        Examples:
            >>> rels_filter, rels_params = self.branch.get_query_filter_path(at=self.at)
            >>> self.params.update(rels_params)
            >>> query += "\n WHERE all(r IN relationships(p) WHERE %s)" % rels_filter

            There is a currently an assumption that the relationship in the path will be named 'r'
        """
        pp = params_prefix
        params: dict[str, Any] = {}
        at = Timestamp(at)
        at_str = at.to_string()
        if branch_agnostic:
            filter_str = f"{variable_name}.from <= ${pp}time1 AND ({variable_name}.to IS NULL or {variable_name}.to >= ${pp}time1)"
            params[f"{pp}time1"] = at_str
            return filter_str, params

        branches_times = self.get_branches_and_times_to_query_global(at=at_str, is_isolated=is_isolated)

        for idx, (branch_name, time_to_query) in enumerate(branches_times.items()):
            params[f"{pp}branch{idx}"] = list(branch_name)
            params[f"{pp}time{idx}"] = time_to_query

        filters = []
        for idx in range(len(branches_times)):
            filters.append(
                f"({variable_name}.branch IN ${pp}branch{idx} AND {variable_name}.from <= ${pp}time{idx} AND {variable_name}.to IS NULL)"
            )
            filters.append(
                f"({variable_name}.branch IN ${pp}branch{idx} AND {variable_name}.from <= ${pp}time{idx} AND {variable_name}.to >= ${pp}time{idx})"
            )

        filter_str = "(" + "\n OR ".join(filters) + ")"

        return filter_str, params

    def get_query_filter_relationships_range(
        self,
        rel_labels: list,
        start_time: Union[Timestamp, str],
        end_time: Union[Timestamp, str],
        include_outside_parentheses: bool = False,
        include_global: bool = False,
    ) -> tuple[list, dict]:
        """Generate a CYPHER Query filter based on a list of relationships to query a range of values in the graph.
        The goal is to return all the values that are valid during this timerange.
        """

        filters = []
        params = {}

        if not isinstance(rel_labels, list):
            raise TypeError(f"rel_labels must be a list, not a {type(rel_labels)}")

        start_time = Timestamp(start_time)
        end_time = Timestamp(end_time)

        if include_global:
            branches_times = self.get_branches_and_times_to_query_global(at=start_time)
        else:
            branches_times = self.get_branches_and_times_to_query(at=start_time)

        params["branches"] = list({branch for branches in branches_times for branch in branches})
        params["start_time"] = start_time.to_string()
        params["end_time"] = end_time.to_string()

        for rel in rel_labels:
            filters_per_rel = [
                f"({rel}.branch IN $branches AND {rel}.from <= $end_time AND {rel}.to IS NULL)",
                f"({rel}.branch IN $branches AND ({rel}.from <= $end_time OR ({rel}.to >= $start_time AND {rel}.to <= $end_time)))",
            ]

            if not include_outside_parentheses:
                filters.append("\n OR ".join(filters_per_rel))

            filters.append("(" + "\n OR ".join(filters_per_rel) + ")")

        return filters, params

    def get_query_filter_relationships_diff(
        self, rel_labels: list, diff_from: Timestamp, diff_to: Timestamp
    ) -> tuple[list, dict]:
        """
        Generate a CYPHER Query filter to query all events that are applicable to a given branch based
        - The time when the branch as created
        - The branched_from time of the branch
        - The diff_to and diff_from time as provided
        """

        if not isinstance(rel_labels, list):
            raise TypeError(f"rel_labels must be a list, not a {type(rel_labels)}")

        start_times, end_times = self.get_branches_and_times_for_range(start_time=diff_from, end_time=diff_to)

        filters = []
        params = {}

        for idx, branch_name in enumerate(start_times.keys()):
            params[f"branch{idx}"] = branch_name
            params[f"start_time{idx}"] = start_times[branch_name]
            params[f"end_time{idx}"] = end_times[branch_name]

        for rel in rel_labels:
            filters_per_rel = []
            for idx in range(len(start_times)):
                filters_per_rel.extend(
                    [
                        f"""({rel}.branch = $branch{idx}
                             AND {rel}.from >= $start_time{idx}
                             AND {rel}.from <= $end_time{idx}
                             AND ( r2.to is NULL or r2.to >= $end_time{idx}))""",
                        f"""({rel}.branch = $branch{idx} AND {rel}.from >= $start_time{idx}
                            AND {rel}.to <= $start_time{idx})""",
                    ]
                )

            filters.append("(" + "\n OR ".join(filters_per_rel) + ")")

        return filters, params

    def get_query_filter_range(
        self, rel_label: list, start_time: Union[Timestamp, str], end_time: Union[Timestamp, str]
    ) -> tuple[list, dict]:
        """
        Generate a CYPHER Query filter to query a range of values in the graph between start_time and end_time."""

        filters = []
        params = {}

        start_time = Timestamp(start_time)
        end_time = Timestamp(end_time)

        params["branches"] = self.get_branches_in_scope()
        params["start_time"] = start_time.to_string()
        params["end_time"] = end_time.to_string()

        filters_per_rel = [
            f"""({rel_label}.branch IN $branches AND {rel_label}.from >= $start_time
                 AND {rel_label}.from <= $end_time AND {rel_label}.to IS NULL)""",
            f"""({rel_label}.branch IN $branches AND (({rel_label}.from >= $start_time
                 AND {rel_label}.from <= $end_time) OR ({rel_label}.to >= $start_time
                 AND {rel_label}.to <= $end_time)))""",
        ]

        filters.append("(" + "\n OR ".join(filters_per_rel) + ")")

        return filters, params

    async def rebase(self, db: InfrahubDatabase, at: Optional[Union[str, Timestamp]] = None) -> None:
        """Rebase the current Branch with its origin branch"""

        at = Timestamp(at)

        # Find all relationships with the name of the branch
        # Delete all relationship that have a to date defined in the past
        # Update the from time on all other relationships
        # If conflict is set, ignore the one with Drop

        await self.rebase_graph(db=db, at=at)

        # FIXME, we must ensure that there is no conflict before rebasing a branch
        #   Otherwise we could endup with a complicated situation
        self.branched_from = at.to_string()
        self.status = BranchStatus.OPEN
        await self.save(db=db)

        # Update the branch in the registry after the rebase
        registry.branch[self.name] = self

    async def rebase_graph(self, db: InfrahubDatabase, at: Optional[Timestamp] = None) -> None:
        at = Timestamp(at)

        query = await GetAllBranchInternalRelationshipQuery.init(db=db, branch=self)
        await query.execute(db=db)

        rels_to_delete = []
        rels_to_update = []
        for result in query.get_results():
            element_id = result.get("r").element_id

            conflict_status = result.get("r").get("conflict", None)
            if conflict_status and conflict_status == "drop":
                rels_to_delete.append(element_id)
                continue

            time_to_str = result.get("r").get("to", None)
            time_from_str = result.get("r").get("from")
            time_from = Timestamp(time_from_str)

            if not time_to_str and time_from_str and time_from <= at:
                rels_to_update.append(element_id)
                continue

            if not time_to_str and time_from_str and time_from > at:
                rels_to_delete.append(element_id)
                continue

            time_to = Timestamp(time_to_str)
            if time_to < at:
                rels_to_delete.append(element_id)
                continue

            rels_to_update.append(element_id)

        update_query = await RebaseBranchUpdateRelationshipQuery.init(db=db, ids=rels_to_update, at=at)
        await update_query.execute(db=db)

        delete_query = await RebaseBranchDeleteRelationshipQuery.init(db=db, ids=rels_to_delete, at=at)
        await delete_query.execute(db=db)


registry.branch_object = Branch



from __future__ import annotations

import re
import warnings
from datetime import datetime, timezone
from typing import Literal, TypedDict

from typing_extensions import NotRequired
from whenever import Date, Instant, LocalDateTime, OffsetDateTime, Time, ZonedDateTime

from .exceptions import TimestampFormatError

UTC = timezone.utc  # Required for older versions of Python


class SubstractParams(TypedDict):
    seconds: NotRequired[float]
    minutes: NotRequired[float]
    hours: NotRequired[float]


REGEX_MAPPING = {
    "seconds": r"(\d+)(s|sec|second|seconds)",
    "minutes": r"(\d+)(m|min|minute|minutes)",
    "hours": r"(\d+)(h|hour|hours)",
}


class Timestamp:
    _obj: ZonedDateTime

    def __init__(self, value: str | ZonedDateTime | Timestamp | None = None) -> None:
        if value and isinstance(value, ZonedDateTime):
            self._obj = value
        elif value and isinstance(value, self.__class__):
            self._obj = value._obj
        elif isinstance(value, str):
            self._obj = self._parse_string(value)
        else:
            self._obj = ZonedDateTime.now("UTC").round(unit="microsecond")

    @property
    def obj(self) -> ZonedDateTime:
        warnings.warn(
            "Direct access to obj property is deprecated. Use to_string(), to_timestamp(), or to_datetime() instead.",
            UserWarning,
            stacklevel=2,
        )
        return self._obj

    @classmethod
    def _parse_string(cls, value: str) -> ZonedDateTime:
        try:
            return ZonedDateTime.parse_common_iso(value)
        except ValueError:
            pass

        try:
            instant_date = Instant.parse_common_iso(value)
            return instant_date.to_tz("UTC")
        except ValueError:
            pass

        try:
            local_date_time = LocalDateTime.parse_common_iso(value)
            return local_date_time.assume_utc().to_tz("UTC")
        except ValueError:
            pass

        try:
            offset_date_time = OffsetDateTime.parse_common_iso(value)
            return offset_date_time.to_tz("UTC")
        except ValueError:
            pass

        try:
            date = Date.parse_common_iso(value)
            local_date = date.at(Time(12, 00))
            return local_date.assume_tz("UTC", disambiguate="compatible")
        except ValueError:
            pass

        params: SubstractParams = {}
        for key, regex in REGEX_MAPPING.items():
            match = re.search(regex, value)
            if match:
                if key == "seconds":
                    params["seconds"] = float(match.group(1))
                elif key == "minutes":
                    params["minutes"] = float(match.group(1))
                elif key == "hours":
                    params["hours"] = float(match.group(1))

        if params:
            return ZonedDateTime.now("UTC").subtract(**params)

        raise TimestampFormatError(f"Invalid time format for {value}")

    def __repr__(self) -> str:
        return f"Timestamp: {self.to_string()}"

    def to_string(self, with_z: bool = True) -> str:
        time_str = self.to_datetime().isoformat(timespec="microseconds")
        if with_z and time_str.endswith("+00:00"):
            time_str = time_str[:-6]
            time_str += "Z"
        return time_str

    def to_timestamp(self) -> int:
        return self._obj.timestamp()

    def to_datetime(self) -> datetime:
        return self._obj.py_datetime()

    def get_obj(self) -> ZonedDateTime:
        return self._obj

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Timestamp):
            return NotImplemented
        return self._obj == other._obj

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, Timestamp):
            return NotImplemented
        return self._obj < other._obj

    def __gt__(self, other: object) -> bool:
        if not isinstance(other, Timestamp):
            return NotImplemented
        return self._obj > other._obj

    def __le__(self, other: object) -> bool:
        if not isinstance(other, Timestamp):
            return NotImplemented
        return self._obj <= other._obj

    def __ge__(self, other: object) -> bool:
        if not isinstance(other, Timestamp):
            return NotImplemented
        return self._obj >= other._obj

    def __hash__(self) -> int:
        return hash(self.to_string())

    def add_delta(self, hours: int = 0, minutes: int = 0, seconds: int = 0, microseconds: int = 0) -> Timestamp:
        warnings.warn(
            "add_delta() is deprecated. Use add() instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.add(hours=hours, minutes=minutes, seconds=seconds, microseconds=microseconds)

    def add(
        self,
        years: int = 0,
        months: int = 0,
        weeks: int = 0,
        days: int = 0,
        hours: float = 0,
        minutes: float = 0,
        seconds: float = 0,
        milliseconds: float = 0,
        microseconds: float = 0,
        nanoseconds: int = 0,
        disambiguate: Literal["compatible"] = "compatible",
    ) -> Timestamp:
        return self.__class__(
            self._obj.add(
                years=years,
                months=months,
                weeks=weeks,
                days=days,
                hours=hours,
                minutes=minutes,
                seconds=seconds,
                milliseconds=milliseconds,
                microseconds=microseconds,
                nanoseconds=nanoseconds,
                disambiguate=disambiguate,
            )
        )

    def subtract(
        self,
        years: int = 0,
        months: int = 0,
        weeks: int = 0,
        days: int = 0,
        hours: float = 0,
        minutes: float = 0,
        seconds: float = 0,
        milliseconds: float = 0,
        microseconds: float = 0,
        nanoseconds: int = 0,
        disambiguate: Literal["compatible"] = "compatible",
    ) -> Timestamp:
        return self.__class__(
            self._obj.subtract(
                years=years,
                months=months,
                weeks=weeks,
                days=days,
                hours=hours,
                minutes=minutes,
                seconds=seconds,
                milliseconds=milliseconds,
                microseconds=microseconds,
                nanoseconds=nanoseconds,
                disambiguate=disambiguate,
            )
        )



from __future__ import annotations

from prefect import flow

from infrahub.context import InfrahubContext  # noqa: TC001  needed for prefect flow
from infrahub.core import registry
from infrahub.core.branch import Branch
from infrahub.core.diff.coordinator import DiffCoordinator
from infrahub.core.diff.diff_locker import DiffLocker
from infrahub.core.diff.merger.merger import DiffMerger
from infrahub.core.diff.repository.repository import DiffRepository
from infrahub.core.merge import BranchMerger
from infrahub.core.validators.determiner import ConstraintValidatorDeterminer
from infrahub.core.validators.models.validate_migration import SchemaValidateMigrationData
from infrahub.core.validators.tasks import schema_validate_migrations
from infrahub.dependencies.registry import get_component_registry
from infrahub.exceptions import ValidationError
from infrahub.workers.dependencies import get_database, get_workflow
from infrahub.workflows.catalogue import BRANCH_MERGE
from infrahub.workflows.utils import add_tags


@flow(name="merge-branch-mutation", flow_run_name="Merge branch graphQL mutation")
async def merge_branch_mutation(branch: str, context: InfrahubContext) -> None:
    await add_tags(branches=[branch])

    database = await get_database()
    async with database.start_session() as db:
        obj = await Branch.get_by_name(db=db, name=branch)
        base_branch = await Branch.get_by_name(db=db, name=registry.default_branch)

        component_registry = get_component_registry()
        diff_coordinator = await component_registry.get_component(DiffCoordinator, db=db, branch=obj)
        diff_repository = await component_registry.get_component(DiffRepository, db=db, branch=obj)
        diff_merger = await component_registry.get_component(DiffMerger, db=db, branch=obj)
        enriched_diff_metadata = await diff_coordinator.update_branch_diff(base_branch=base_branch, diff_branch=obj)
        async for _ in diff_repository.get_all_conflicts_for_diff(
            diff_branch_name=enriched_diff_metadata.diff_branch_name, diff_id=enriched_diff_metadata.uuid
        ):
            # if there are any conflicts, raise the error
            raise ValidationError(
                f"Branch {obj.name} contains conflicts with the default branch."
                " Please create a Proposed Change to resolve the conflicts or manually update them before merging."
            )
        node_diff_field_summaries = await diff_repository.get_node_field_summaries(
            diff_branch_name=enriched_diff_metadata.diff_branch_name, diff_id=enriched_diff_metadata.uuid
        )

        merger = BranchMerger(
            db=db,
            diff_coordinator=diff_coordinator,
            diff_merger=diff_merger,
            diff_repository=diff_repository,
            source_branch=obj,
            diff_locker=DiffLocker(),
            workflow=get_workflow(),
        )
        candidate_schema = merger.get_candidate_schema()
        determiner = ConstraintValidatorDeterminer(schema_branch=candidate_schema)
        constraints = await determiner.get_constraints(node_diffs=node_diff_field_summaries)
        if obj.has_schema_changes:
            constraints += await merger.calculate_validations(target_schema=candidate_schema)

        if constraints:
            responses = await schema_validate_migrations(
                message=SchemaValidateMigrationData(branch=obj, schema_branch=candidate_schema, constraints=constraints)
            )
            error_messages = [violation.message for response in responses for violation in response.violations]
            if error_messages:
                raise ValidationError(",\n".join(error_messages))

        await get_workflow().execute_workflow(workflow=BRANCH_MERGE, context=context, parameters={"branch": obj.name})



from graphene import Enum

from infrahub.core import constants
from infrahub.core.branch.enums import BranchStatus
from infrahub.permissions import constants as permission_constants

CheckType = Enum.from_enum(constants.CheckType)

DiffAction = Enum.from_enum(constants.DiffAction)

Severity = Enum.from_enum(constants.Severity)

BranchRelativePermissionDecision = Enum.from_enum(permission_constants.BranchRelativePermissionDecision)

InfrahubBranchStatus = Enum.from_enum(BranchStatus)



import importlib
import inspect
from typing import Any, Awaitable, TypeVar
from uuid import UUID

from prefect import Flow
from prefect.client.orchestration import PrefectClient
from prefect.client.schemas.actions import DeploymentScheduleCreate
from prefect.client.schemas.objects import ConcurrencyLimitStrategy, FlowRun
from prefect.client.schemas.schedules import CronSchedule
from pydantic import BaseModel, Field
from typing_extensions import Self

from infrahub import __version__

from .constants import TAG_NAMESPACE, WorkflowTag, WorkflowType

TASK_RESULT_STORAGE_NAME = "infrahub-storage"

WorkflowReturn = TypeVar("WorkflowReturn")


class WorkerPoolDefinition(BaseModel):
    name: str
    worker_type: str | None = None
    description: str = ""


class WorkflowInfo(BaseModel):
    id: UUID
    info: FlowRun | None = None

    @classmethod
    def from_flow(cls, flow_run: FlowRun) -> Self:
        return cls(id=flow_run.id, info=flow_run)


class WorkflowParameter(BaseModel):
    name: str
    type: str
    required: bool


class WorkflowDefinition(BaseModel):
    name: str
    type: WorkflowType = WorkflowType.INTERNAL
    module: str
    function: str
    cron: str | None = None
    tags: list[WorkflowTag] = Field(default_factory=list)
    concurrency_limit: int | None = Field(
        default=None,
        description="The concurrency limit for the deployment.",
    )
    concurrency_limit_strategy: ConcurrencyLimitStrategy | None = Field(
        default=None,
        description="The concurrency options for the deployment.",
    )

    @property
    def entrypoint(self) -> str:
        if self.type == WorkflowType.USER:
            return f"{self.module}:{self.function}"
        return f"backend/{self.module.replace('.', '/')}:{self.function}"

    @property
    def full_name(self) -> str:
        return f"{self.name}/{self.name}"

    def to_deployment(self) -> dict[str, Any]:
        payload: dict[str, Any] = {
            "name": self.name,
            "entrypoint": self.entrypoint,
            "tags": self.get_tags(),
            "concurrency_limit": self.concurrency_limit,
        }
        if self.concurrency_limit_strategy:
            payload["concurrency_options"] = {"collision_strategy": self.concurrency_limit_strategy}
        if self.type == WorkflowType.CORE:
            payload["version"] = __version__
        if self.cron:
            payload["schedules"] = [DeploymentScheduleCreate(schedule=CronSchedule(cron=self.cron))]

        return payload

    def get_tags(self) -> list[str]:
        tags: list[str] = []
        if self.type != WorkflowType.INTERNAL:
            tags.append(TAG_NAMESPACE)
        tags.append(WorkflowTag.WORKFLOWTYPE.render(identifier=self.type.value))
        tags += [tag.render() for tag in self.tags]
        return tags

    async def save(self, client: PrefectClient, work_pool: WorkerPoolDefinition) -> UUID:
        flow_id = await client.create_flow_from_name(self.name)
        data = self.to_deployment()
        data["work_pool_name"] = work_pool.name
        return await client.create_deployment(flow_id=flow_id, **data)

    def load_function(self) -> Flow[Any, Awaitable]:
        module = importlib.import_module(self.module)
        return getattr(module, self.function)

    def get_parameters(self) -> dict[str, WorkflowParameter]:
        fn = self.load_function()
        signature = inspect.signature(fn.fn)
        required_params = fn.parameters.required

        ANNOTATIONS_TO_EXCLUDE = ["InfrahubServices"]

        params = {}
        for item in signature.parameters.values():
            # Workflow signature will return a string if 'from __future__ import annotations' is used
            # Otherwise it will return a type object
            annotation = item.annotation if isinstance(item.annotation, str) else item.annotation.__name__
            if annotation in ANNOTATIONS_TO_EXCLUDE:
                continue
            param = WorkflowParameter(name=item.name, type=annotation, required=item.name in required_params)
            params[item.name] = param

        return params



from __future__ import annotations

import ipaddress
from typing import TYPE_CHECKING, Any

from graphql.type.definition import GraphQLNonNull
from netaddr import IPSet
from opentelemetry import trace

from infrahub.core import registry
from infrahub.core.constants import InfrahubKind
from infrahub.core.ipam.constants import PrefixMemberType
from infrahub.core.manager import NodeManager
from infrahub.core.node import Node
from infrahub.core.protocols import BuiltinIPNamespace, BuiltinIPPrefix
from infrahub.core.schema.generic_schema import GenericSchema
from infrahub.exceptions import ValidationError
from infrahub.graphql.parser import extract_selection
from infrahub.graphql.permissions import get_permissions

from ..models import OrderModel

if TYPE_CHECKING:
    from collections.abc import Sequence

    from graphql import GraphQLResolveInfo
    from pydantic import IPvAnyAddress, IPvAnyInterface, IPvAnyNetwork

    from infrahub.core.branch.models import Branch
    from infrahub.core.schema import NodeSchema
    from infrahub.database import InfrahubDatabase
    from infrahub.graphql.initialization import GraphqlContext
    from infrahub.graphql.models import OrderModel


def _ip_range_display_label(node: Node) -> str:
    """Return a human friendly summary of an IP range"""
    size = int(node.last_address.obj) - int(node.address.obj) + 1

    if size == 1:
        return "1 IP address available"
    if size <= 2**16:
        return f"{size} IP addresses available"
    return f"More than {2**16} IP addresses available"


def _ip_with_prefix_length(ip_address: IPvAnyAddress, ip_prefix: IPvAnyNetwork) -> IPvAnyInterface:
    """Convert an `IPAddress` object into an `IPInterface` one given a `IPNetwork`."""
    return ipaddress.ip_interface(f"{ip_address}/{ip_prefix.prefixlen}")


async def _build_ip_range_node(
    db: InfrahubDatabase,
    branch: Branch,
    schema: NodeSchema,
    address: IPvAnyAddress,
    last_address: IPvAnyAddress,
    ip_namespace: BuiltinIPNamespace,
    ip_prefix: BuiltinIPPrefix,
) -> Node:
    address_with_len = str(_ip_with_prefix_length(ip_address=address, ip_prefix=ip_prefix.prefix.obj))
    last_address_with_len = str(_ip_with_prefix_length(ip_address=last_address, ip_prefix=ip_prefix.prefix.obj))

    n = await Node.init(schema=schema, db=db, branch=branch)
    await n.new(
        db=db,
        address=address_with_len,
        last_address=last_address_with_len,
        description=f"Available IP range {address_with_len} - {last_address_with_len}",
        ip_namespace=ip_namespace,
        ip_prefix=ip_prefix,
    )
    return n


def _include_first_and_last_ips(ip_prefix: BuiltinIPPrefix) -> bool:
    if ip_prefix.prefix.version == 6 or ip_prefix.is_pool.value:
        return True

    return ip_prefix.member_type.value == PrefixMemberType.ADDRESS.value and ip_prefix.prefix.prefixlen == 31


async def _resolve_available_address_nodes(
    db: InfrahubDatabase,
    branch: Branch,
    prefix: BuiltinIPPrefix,
    existing_nodes: Sequence[Node],
    first_node_context: Node | None = None,
    last_node_context: Node | None = None,
) -> list[Node]:
    """Annotate a list of IP addresses node with available ranges within a prefix."""
    ip_prefix: IPvAnyNetwork = prefix.prefix.obj
    ip_namespace = await prefix.ip_namespace.get_peer(db=db, peer_type=BuiltinIPNamespace, raise_on_error=True)
    ip_range_schema = registry.get_node_schema(name=InfrahubKind.IPRANGEAVAILABLE, branch=branch)

    # Make sure nodes are ordered by addresses
    sorted_nodes = sorted(existing_nodes, key=lambda n: n.address.obj)
    prefix_first_address = (
        ip_prefix.network_address if _include_first_and_last_ips(ip_prefix=prefix) else ip_prefix.network_address + 1
    )
    prefix_last_address = (
        ip_prefix.broadcast_address
        if _include_first_and_last_ips(ip_prefix=prefix)
        else ip_prefix.broadcast_address - 1
    )

    if not sorted_nodes:
        return [
            await _build_ip_range_node(
                db=db,
                branch=branch,
                schema=ip_range_schema,
                address=prefix_first_address,
                last_address=prefix_last_address,
                ip_namespace=ip_namespace,
                ip_prefix=prefix,
            )
        ]

    first_address: IPvAnyAddress = prefix_first_address
    last_address: IPvAnyAddress = prefix_last_address

    # Use but exclude context addresses to avoid having them in the result
    if first_node_context:
        first_address = first_node_context.address.obj.ip + 1
    if last_node_context:
        last_address = last_node_context.address.obj.ip - 1

    with_available_ranges: list[Node] = []
    previous_address: IPvAnyAddress | None = None

    # Look for a gap at the beginning of the prefix
    if sorted_nodes[0].address.obj.ip > first_address:
        with_available_ranges.append(
            await _build_ip_range_node(
                db=db,
                branch=branch,
                schema=ip_range_schema,
                address=first_address,
                last_address=sorted_nodes[0].address.obj.ip - 1,
                ip_namespace=ip_namespace,
                ip_prefix=prefix,
            )
        )

    # Look for gaps between existing addresses
    for existing in sorted_nodes:
        current = existing.address.obj.ip
        if previous_address:
            if int(current) - int(previous_address) > 1:
                with_available_ranges.append(
                    await _build_ip_range_node(
                        db=db,
                        branch=branch,
                        schema=ip_range_schema,
                        address=previous_address + 1,
                        last_address=current - 1,
                        ip_namespace=ip_namespace,
                        ip_prefix=prefix,
                    )
                )

        with_available_ranges.append(existing)
        previous_address = existing.address.obj.ip

    # Look for a gap at the end of the prefix
    if previous_address and previous_address < last_address:
        with_available_ranges.append(
            await _build_ip_range_node(
                db=db,
                branch=branch,
                schema=ip_range_schema,
                address=previous_address + 1,
                last_address=last_address,
                ip_namespace=ip_namespace,
                ip_prefix=prefix,
            )
        )

    return with_available_ranges


async def _resolve_available_prefix_nodes(
    db: InfrahubDatabase,
    branch: Branch,
    prefix: BuiltinIPPrefix,
    existing_nodes: Sequence[Node],
    first_node_context: Node | None = None,
    last_node_context: Node | None = None,
) -> list[Node]:
    """Annotate a list of IP prefixes node with available prefixes within a parent one."""
    ip_prefix_schema = registry.get_node_schema(name=InfrahubKind.IPPREFIXAVAILABLE, branch=branch)

    existing_prefixes = IPSet([n.prefix.value for n in existing_nodes])
    if first_node_context:
        existing_prefixes.add(first_node_context.prefix.value)
    if last_node_context:
        existing_prefixes.add(last_node_context.prefix.value)

    # Infer which prefixes are actually available
    available_prefixes = IPSet([prefix.prefix.value]) ^ existing_prefixes
    available_nodes: list[Node] = []

    # Turn them into nodes (without saving them in the database)
    for available_prefix in available_prefixes.iter_cidrs():
        p = ipaddress.ip_network(str(available_prefix))
        if (first_node_context and p < first_node_context.prefix.obj) or (
            last_node_context and p > last_node_context.prefix.obj
        ):
            continue

        node = await Node.init(schema=ip_prefix_schema, db=db, branch=branch)
        await node.new(
            db=db, prefix=str(available_prefix), ip_namespace=await prefix.ip_namespace.get_peer(db=db), parent=prefix
        )
        available_nodes.append(node)

    # Properly sort existing nodes with available prefixes
    with_available_prefixes = sorted(existing_nodes + available_nodes, key=lambda n: n.prefix.obj)

    if len(with_available_prefixes) > 1 or with_available_prefixes[0].prefix.obj != prefix.prefix.obj:
        return with_available_prefixes

    # If the only available prefix is the same as the container prefix, this means the container prefix is empty and we should therefore at least
    # offer two smaller prefixes allocatable within it
    available_nodes.clear()

    for subnet in prefix.prefix.obj.subnets():
        node = await Node.init(schema=ip_prefix_schema, db=db, branch=branch)
        await node.new(db=db, prefix=str(subnet), ip_namespace=await prefix.ip_namespace.get_peer(db=db), parent=prefix)
        available_nodes.append(node)

    return available_nodes


def _filter_kinds(nodes: list[Node], kinds: list[str], limit: int | None) -> list[Node]:
    filtered: list[Node] = []
    available_node_kinds = [InfrahubKind.IPPREFIXAVAILABLE, InfrahubKind.IPRANGEAVAILABLE]
    kinds_with_available = kinds + available_node_kinds

    limit_with_available = limit
    for node in nodes:
        if node.get_schema().kind not in kinds_with_available:
            continue
        # Adapt the limit of nodes to return by always including available ones
        if limit and node.get_schema().kind in available_node_kinds:
            limit_with_available += 1
        filtered.append(node)

    return filtered[:limit_with_available] if limit else filtered


async def _annotate_result(
    db: InfrahubDatabase,
    branch: Branch,
    resolve_available: bool,
    schema: NodeSchema | GenericSchema,
    parent_prefix: BuiltinIPPrefix | None,
    result: list[Node],
    first_node_context: Node | None = None,
    last_node_context: Node | None = None,
    kinds_to_filter: list[str] | None = None,
    limit: int | None = None,
) -> list[Node]:
    nodes: list[Node] = result

    if resolve_available and parent_prefix:
        if schema.is_ip_address:
            nodes = await _resolve_available_address_nodes(
                db=db,
                branch=branch,
                prefix=parent_prefix,
                existing_nodes=result,
                first_node_context=first_node_context,
                last_node_context=last_node_context,
            )
        else:
            nodes = await _resolve_available_prefix_nodes(
                db=db,
                branch=branch,
                prefix=parent_prefix,
                existing_nodes=result,
                first_node_context=first_node_context,
                last_node_context=last_node_context,
            )

    return _filter_kinds(nodes=nodes, kinds=kinds_to_filter, limit=limit) if kinds_to_filter else nodes


@trace.get_tracer(__name__).start_as_current_span("ipam_paginated_list_resolver")
async def ipam_paginated_list_resolver(  # noqa: PLR0915
    root: dict,  # noqa: ARG001
    info: GraphQLResolveInfo,
    offset: int | None = None,
    limit: int | None = None,
    order: OrderModel | None = None,
    partial_match: bool = False,
    **kwargs: dict[str, Any],
) -> dict[str, Any]:
    schema: NodeSchema | GenericSchema = (
        info.return_type.of_type.graphene_type._meta.schema
        if isinstance(info.return_type, GraphQLNonNull)
        else info.return_type.graphene_type._meta.schema
    )

    if not isinstance(schema, GenericSchema) or schema.kind not in [InfrahubKind.IPADDRESS, InfrahubKind.IPPREFIX]:
        raise ValidationError(f"{schema.kind} is not {InfrahubKind.IPADDRESS} or {InfrahubKind.IPPREFIX}")

    fields = await extract_selection(info=info, schema=schema)
    resolve_available = bool(kwargs.pop("include_available", False))
    kinds_to_filter: list[str] = kwargs.pop("kinds", [])  # type: ignore[assignment]

    for kind in kinds_to_filter:
        if kind not in schema.used_by:
            raise ValidationError(f"{kind} is not a node inheriting from {schema.kind}")

    graphql_context: GraphqlContext = info.context
    async with graphql_context.db.start_session(read_only=True) as db:
        response: dict[str, Any] = {"edges": []}
        filters = {
            key: value for key, value in kwargs.items() if ("__" in key and value is not None) or key in ("ids", "hfid")
        }

        edges = fields.get("edges", {})
        node_fields = edges.get("node", {})

        permission_set: dict[str, Any] | None = None
        permissions = (
            await get_permissions(schema=schema, graphql_context=graphql_context)
            if graphql_context.permissions
            else None
        )
        if fields.get("permissions"):
            response["permissions"] = permissions

        if permissions:
            for edge in permissions["edges"]:
                if edge["node"]["kind"] == schema.kind:
                    permission_set = edge["node"]

        parent_prefix_id = ""
        if schema.is_ip_address and "ip_prefix__ids" in filters:
            parent_prefix_id = next(iter(filters["ip_prefix__ids"]))
        if schema.is_ip_prefix and "parent__ids" in filters:
            parent_prefix_id = next(iter(filters["parent__ids"]))

        parent_prefix: BuiltinIPPrefix | None = None
        if parent_prefix_id:
            parent_prefix = await NodeManager.get_one(
                db=db, kind=BuiltinIPPrefix, id=parent_prefix_id, at=graphql_context.at, branch=graphql_context.branch
            )

        first_node_context: Node | None = None
        fetch_first_node_context = False
        if offset is not None and offset > 0:
            offset -= 1
            fetch_first_node_context = True

        last_node_context: Node | None = None
        fetch_last_node_context = False
        if limit is not None and limit > 0:
            limit += 1
            fetch_last_node_context = True

        # Since we are going to narrow down the number of nodes in the end, we will query for a larger set (that can potentially include all kinds of
        # implementations) in the first place to make sure that we will fill in the page to its maximum
        query_limit = limit
        if kinds_to_filter and limit:
            query_limit *= len(schema.used_by)

        objs = []
        if edges or "hfid" in filters:
            objs = await NodeManager.query(
                db=db,
                schema=schema,
                filters=filters or None,
                fields=node_fields,
                at=graphql_context.at,
                branch=graphql_context.branch,
                limit=query_limit,
                offset=offset,
                account=graphql_context.account_session,
                include_source=True,
                include_owner=True,
                partial_match=partial_match,
                order=order,
            )

            if fetch_first_node_context and len(objs) > 2:
                first_node_context = objs[0]
                objs = objs[1:]
            if fetch_last_node_context and len(objs) >= limit >= 2:
                last_node_context = objs[-1]
                objs = objs[:-1]

        if "count" in fields:
            if filters.get("hfid"):
                response["count"] = len(objs)
            else:
                response["count"] = await NodeManager.count(
                    db=db,
                    schema=schema,
                    filters=filters,
                    at=graphql_context.at,
                    branch=graphql_context.branch,
                    partial_match=partial_match,
                )

        result = await _annotate_result(
            db=db,
            branch=graphql_context.branch,
            resolve_available=resolve_available,
            schema=schema,
            parent_prefix=parent_prefix,
            result=objs,
            first_node_context=first_node_context,
            last_node_context=last_node_context,
            kinds_to_filter=kinds_to_filter,
            limit=limit,
        )

        if result:
            objects = []
            for obj in result:
                obj_data = await obj.to_graphql(
                    db=db,
                    fields=node_fields,
                    related_node_ids=graphql_context.related_node_ids,
                    permissions=permission_set,
                )

                # Override display label for available IP ranges
                if obj.get_schema().kind == InfrahubKind.IPRANGEAVAILABLE and "display_label" in obj_data:
                    obj_data["display_label"] = _ip_range_display_label(node=obj)

                objects.append({"node": obj_data})

            response["edges"] = objects

        return response



from infrahub import config
from infrahub.auth import AccountSession
from infrahub.core import registry
from infrahub.core.account import GlobalPermission, ObjectPermission
from infrahub.core.branch import Branch
from infrahub.core.constants import GLOBAL_BRANCH_NAME, GlobalPermissions, InfrahubKind, PermissionDecision
from infrahub.core.manager import get_schema
from infrahub.core.schema.node_schema import NodeSchema
from infrahub.database import InfrahubDatabase
from infrahub.graphql.analyzer import GraphQLOperation, InfrahubGraphQLQueryAnalyzer
from infrahub.graphql.initialization import GraphqlParams
from infrahub.permissions.constants import PermissionDecisionFlag
from infrahub.utils import extract_camelcase_words

from .interface import CheckerResolution, GraphQLQueryPermissionCheckerInterface


class ObjectPermissionChecker(GraphQLQueryPermissionCheckerInterface):
    """Checker that makes sure a user account can perform some action on some kind of objects."""

    async def supports(self, db: InfrahubDatabase, account_session: AccountSession, branch: Branch) -> bool:  # noqa: ARG002
        return config.SETTINGS.main.allow_anonymous_access or account_session.authenticated

    async def check(
        self,
        db: InfrahubDatabase,  # noqa: ARG002
        account_session: AccountSession,  # noqa: ARG002
        analyzed_query: InfrahubGraphQLQueryAnalyzer,
        query_parameters: GraphqlParams,
        branch: Branch,  # noqa: ARG002
    ) -> CheckerResolution:
        required_decision = (
            PermissionDecisionFlag.ALLOW_DEFAULT
            if analyzed_query.branch is None
            or analyzed_query.branch.name in (GLOBAL_BRANCH_NAME, registry.default_branch)
            else PermissionDecisionFlag.ALLOW_OTHER
        )

        permissions: list[ObjectPermission] = []
        for kind, object_access in analyzed_query.query_report.requested_read.items():
            if object_access.attributes or object_access.relationships:
                extracted_words = extract_camelcase_words(kind)
                permissions.append(
                    ObjectPermission(
                        namespace=extracted_words[0],
                        name="".join(extracted_words[1:]),
                        action="view",
                        decision=required_decision,
                    )
                )

        for kind, requested_permissions in analyzed_query.query_report.kind_action_map.items():
            for requested_permission in requested_permissions:
                extracted_words = extract_camelcase_words(kind)
                permissions.append(
                    ObjectPermission(
                        namespace=extracted_words[0],
                        name="".join(extracted_words[1:]),
                        action=requested_permission.value,
                        decision=required_decision,
                    )
                )

        query_parameters.context.active_permissions.raise_for_permissions(permissions=permissions)

        return CheckerResolution.TERMINATE


class AccountManagerPermissionChecker(GraphQLQueryPermissionCheckerInterface):
    """Checker that makes sure a user account can perform actions on account related objects.

    This is similar to object permission checker except that we care for any operations on any account related kinds.
    """

    permission_required = GlobalPermission(
        action=GlobalPermissions.MANAGE_ACCOUNTS.value, decision=PermissionDecision.ALLOW_ALL.value
    )

    async def supports(self, db: InfrahubDatabase, account_session: AccountSession, branch: Branch) -> bool:  # noqa: ARG002
        return config.SETTINGS.main.allow_anonymous_access or account_session.authenticated

    async def check(
        self,
        db: InfrahubDatabase,
        account_session: AccountSession,  # noqa: ARG002
        analyzed_query: InfrahubGraphQLQueryAnalyzer,
        query_parameters: GraphqlParams,
        branch: Branch,
    ) -> CheckerResolution:
        is_account_operation = False
        kinds = analyzed_query.query_report.impacted_models
        operation_names = [operation.name for operation in analyzed_query.operations]

        for kind in kinds:
            schema = get_schema(db=db, branch=branch, node_schema=kind)
            if is_account_operation := kind in (
                InfrahubKind.GENERICACCOUNT,
                InfrahubKind.ACCOUNTGROUP,
                InfrahubKind.ACCOUNTROLE,
            ) or (isinstance(schema, NodeSchema) and InfrahubKind.GENERICACCOUNT in schema.inherit_from):
                break

        # Ignore non-account related operation or viewing account own profile
        if not is_account_operation or operation_names == ["AccountProfile"]:
            return CheckerResolution.NEXT_CHECKER

        if analyzed_query.contains_mutation:
            query_parameters.context.active_permissions.raise_for_permission(permission=self.permission_required)

        return CheckerResolution.NEXT_CHECKER


class PermissionManagerPermissionChecker(GraphQLQueryPermissionCheckerInterface):
    """Checker that makes sure a user account can perform actions on permission related object.

    This is similar to object permission checker except that we care for any operations on any permission related kinds.
    """

    permission_required = GlobalPermission(
        action=GlobalPermissions.MANAGE_PERMISSIONS.value, decision=PermissionDecision.ALLOW_ALL.value
    )
    # Map kinds and the relationship to protect from being read
    kind_relationship_to_check = {
        InfrahubKind.ACCOUNTROLE: "permissions",
        InfrahubKind.BASEPERMISSION: "roles",
        InfrahubKind.GLOBALPERMISSION: "roles",
        InfrahubKind.OBJECTPERMISSION: "roles",
    }

    async def supports(self, db: InfrahubDatabase, account_session: AccountSession, branch: Branch) -> bool:  # noqa: ARG002
        return config.SETTINGS.main.allow_anonymous_access or account_session.authenticated

    async def check(
        self,
        db: InfrahubDatabase,  # noqa: ARG002
        account_session: AccountSession,  # noqa: ARG002
        analyzed_query: InfrahubGraphQLQueryAnalyzer,
        query_parameters: GraphqlParams,
        branch: Branch,  # noqa: ARG002
    ) -> CheckerResolution:
        for kind, relationship in self.kind_relationship_to_check.items():
            if (
                kind in analyzed_query.query_report.requested_read
                and relationship in analyzed_query.query_report.requested_read[kind].relationships
            ):
                query_parameters.context.active_permissions.raise_for_permission(permission=self.permission_required)

        for query in analyzed_query.query_report.queries:
            if not query.infrahub_model:
                continue

            # Prevent mutations on permissions and account roles
            if (
                query.operation == GraphQLOperation.MUTATION
                and isinstance(query.infrahub_model, NodeSchema)
                and (
                    InfrahubKind.BASEPERMISSION in query.infrahub_model.inherit_from
                    or query.infrahub_model.kind == InfrahubKind.ACCOUNTROLE
                )
            ):
                query_parameters.context.active_permissions.raise_for_permission(permission=self.permission_required)

        return CheckerResolution.NEXT_CHECKER


class RepositoryManagerPermissionChecker(GraphQLQueryPermissionCheckerInterface):
    """Checker that makes sure a user account can add/edit/delete repository objects.

    This is similar to object permission checker except that we only care about mutations on repositories.
    """

    permission_required = GlobalPermission(
        action=GlobalPermissions.MANAGE_REPOSITORIES.value, decision=PermissionDecision.ALLOW_ALL.value
    )

    async def supports(self, db: InfrahubDatabase, account_session: AccountSession, branch: Branch) -> bool:  # noqa: ARG002
        return config.SETTINGS.main.allow_anonymous_access or account_session.authenticated

    async def check(
        self,
        db: InfrahubDatabase,
        account_session: AccountSession,  # noqa: ARG002
        analyzed_query: InfrahubGraphQLQueryAnalyzer,
        query_parameters: GraphqlParams,
        branch: Branch,
    ) -> CheckerResolution:
        is_repository_operation = False
        kinds = analyzed_query.query_report.impacted_models

        for kind in kinds:
            schema = get_schema(db=db, branch=branch, node_schema=kind)
            if is_repository_operation := kind in (
                InfrahubKind.GENERICREPOSITORY,
                InfrahubKind.REPOSITORY,
                InfrahubKind.READONLYREPOSITORY,
            ) or (isinstance(schema, NodeSchema) and InfrahubKind.GENERICREPOSITORY in schema.inherit_from):
                break

        if is_repository_operation and analyzed_query.contains_mutation:
            query_parameters.context.active_permissions.raise_for_permission(permission=self.permission_required)

        return CheckerResolution.NEXT_CHECKER



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from infrahub.core.query import Query, QueryType

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.core.timestamp import Timestamp
    from infrahub.database import InfrahubDatabase


class DiffMergeQuery(Query):
    name = "diff_merge"
    type = QueryType.WRITE
    insert_return = False

    def __init__(
        self,
        node_diff_dicts: dict[str, Any],
        at: Timestamp,
        target_branch: Branch,
        migrated_kinds_id_map: dict[str, str],
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.node_diff_dicts = node_diff_dicts
        self.at = at
        self.target_branch = target_branch
        self.source_branch_name = self.branch.name
        self.migrated_kinds_id_map = migrated_kinds_id_map

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        self.params = {
            "node_diff_dicts": self.node_diff_dicts,
            "at": self.at.to_string(),
            "branch_level": self.target_branch.hierarchy_level,
            "target_branch": self.target_branch.name,
            "source_branch": self.source_branch_name,
            "migrated_kinds_id_map": self.migrated_kinds_id_map,
            "migrated_kinds_uuids": list(self.migrated_kinds_id_map.keys()),
        }
        # ruff: noqa: E501
        query = """
UNWIND $node_diff_dicts AS node_diff_map
WITH node_diff_map, node_diff_map.uuid IN $migrated_kinds_uuids AS is_node_kind_migration
WITH node_diff_map, is_node_kind_migration, CASE
    WHEN $migrated_kinds_uuids IS NULL THEN NULL
    WHEN is_node_kind_migration THEN $migrated_kinds_id_map[node_diff_map.uuid]
    ELSE NULL
END AS node_db_id

// ------------------------------
// find the correct Node if the Node had its kind/inheritance migrated
// and there are multiple Nodes with the same UUID
// ------------------------------
CALL (node_diff_map, node_db_id) {
    MATCH (n:Node {uuid: node_diff_map.uuid})-[n_is_part_of:IS_PART_OF]->(:Root)
    WHERE node_db_id IS NULL OR %(id_func)s(n) = node_db_id
    AND n_is_part_of.branch IN [$source_branch, $target_branch]
    RETURN n
    ORDER BY n_is_part_of.branch_level DESC, n_is_part_of.from DESC, n_is_part_of.status ASC
    LIMIT 1
}
WITH n, node_diff_map, is_node_kind_migration
CALL (n, node_diff_map, is_node_kind_migration) {
    WITH CASE
        WHEN node_diff_map.action = "ADDED" THEN "active"
        WHEN node_diff_map.action = "REMOVED" THEN "deleted"
        ELSE NULL
    END AS node_rel_status
    CALL (n, node_diff_map, is_node_kind_migration, node_rel_status) {
        // ------------------------------
        // only make IS_PART_OF updates if node is ADDED or REMOVED
        // ------------------------------
        WITH node_rel_status
        WHERE node_rel_status IS NOT NULL
        // nodes with a migrated kind are handled in DiffMergeMigratedKindsQuery
        AND is_node_kind_migration = FALSE
        MATCH (root:Root)
        // ------------------------------
        // set IS_PART_OF.to, optionally, target branch
        // ------------------------------
        WITH root, n, node_rel_status
        CALL (root, n, node_rel_status) {
            OPTIONAL MATCH (root)<-[target_r_root:IS_PART_OF {branch: $target_branch, status: "active"}]-(n)
            WHERE node_rel_status = "deleted"
            AND target_r_root.from <= $at AND target_r_root.to IS NULL
            SET target_r_root.to = $at
        }
        // ------------------------------
        // create new IS_PART_OF relationship on target_branch
        // ------------------------------
        WITH root, n, node_rel_status
        CALL (root, n, node_rel_status) {
            OPTIONAL MATCH (root)<-[r_root:IS_PART_OF {branch: $target_branch}]-(n)
            WHERE r_root.status = node_rel_status
            AND r_root.from <= $at
            AND (r_root.to >= $at OR r_root.to IS NULL)
            WITH r_root
            WHERE r_root IS NULL
            CREATE (root)
                <-[:IS_PART_OF { branch: $target_branch, branch_level: $branch_level, from: $at, status: node_rel_status }]
                -(n)
        }
        // ------------------------------
        // shortcut to delete all attributes and relationships for this node if the node is deleted
        // ------------------------------
        CALL (n, node_rel_status) {
            WITH n, node_rel_status
            WHERE node_rel_status = "deleted"
            CALL (n) {
                OPTIONAL MATCH (n)-[rel1:IS_RELATED]-(:Relationship)-[rel2]-(p)
                WHERE (p.uuid IS NULL OR n.uuid <> p.uuid)
                AND rel1.branch = $target_branch
                AND rel2.branch = $target_branch
                AND rel1.status = "active"
                AND rel2.status = "active"
                RETURN rel1, rel2
                UNION
                OPTIONAL MATCH (n)-[rel1:HAS_ATTRIBUTE]->(:Attribute)-[rel2]->()
                WHERE type(rel2) <> "HAS_ATTRIBUTE"
                AND rel1.branch = $target_branch
                AND rel2.branch = $target_branch
                AND rel1.status = "active"
                AND rel2.status = "active"
                RETURN rel1, rel2
            }
            WITH rel1, rel2
            WHERE rel1.to IS NULL
            AND rel2.to IS NULL
            AND rel1.from <= $at
            AND rel2.from <= $at
            SET rel1.to = $at
            SET rel2.to = $at
            // ------------------------------
            // and delete HAS_OWNER and HAS_SOURCE edges to this node if the node is deleted
            // ------------------------------
            WITH n
            CALL (n) {
                CALL (n) {
                    MATCH (n)<-[rel:HAS_OWNER]-()
                    WHERE rel.branch = $target_branch
                    AND rel.status = "active"
                    AND rel.from <= $at
                    AND rel.to IS NULL
                    RETURN rel
                    UNION
                    MATCH (n)<-[rel:HAS_SOURCE]-()
                    WHERE rel.branch = $target_branch
                    AND rel.status = "active"
                    AND rel.from <= $at
                    AND rel.to IS NULL
                    RETURN rel
                }
                SET rel.to = $at
            }
        }
    }
    WITH n, node_diff_map
    CALL (n, node_diff_map) {
        WITH CASE
            WHEN node_diff_map.attributes IS NULL OR node_diff_map.attributes = [] THEN [NULL]
            ELSE node_diff_map.attributes
        END AS attribute_maps
        UNWIND attribute_maps AS attribute_diff_map
        // ------------------------------
        // handle updates for attributes under this node
        // ------------------------------
        CALL (n, attribute_diff_map) {
            WITH attribute_diff_map.name AS attr_name, CASE
                WHEN attribute_diff_map.action = "ADDED" THEN "active"
                WHEN attribute_diff_map.action = "REMOVED" THEN "deleted"
                ELSE NULL
            END AS attr_rel_status
            CALL (n, attr_name) {
                OPTIONAL MATCH (n)-[has_attr:HAS_ATTRIBUTE]->(a:Attribute {name: attr_name})
                WHERE has_attr.branch IN [$source_branch, $target_branch]
                RETURN a
                ORDER BY has_attr.from DESC
                LIMIT 1
            }
            WITH n, attr_rel_status, a
            // ------------------------------
            // set HAS_ATTRIBUTE.to on target branch if necessary
            // ------------------------------
            CALL (n, attr_rel_status, a) {
                OPTIONAL MATCH (n)
                    -[target_r_attr:HAS_ATTRIBUTE {branch: $target_branch, status: "active"}]
                    ->(a)
                WHERE attr_rel_status = "deleted"
                AND target_r_attr.from <= $at AND target_r_attr.to IS NULL
                SET target_r_attr.to = $at
            }
            WITH n, attr_rel_status, a
            // ------------------------------
            // conditionally create new HAS_ATTRIBUTE relationship on target_branch, if necessary
            // ------------------------------
            CALL (n, attr_rel_status, a) {
                WITH n, attr_rel_status, a
                WHERE a IS NOT NULL
                OPTIONAL MATCH (n)-[r_attr:HAS_ATTRIBUTE {branch: $target_branch}]->(a)
                WHERE r_attr.status = attr_rel_status
                AND r_attr.from <= $at
                AND (r_attr.to >= $at OR r_attr.to IS NULL)
                WITH r_attr
                WHERE r_attr IS NULL
                CREATE (n)-[:HAS_ATTRIBUTE { branch: $target_branch, branch_level: $branch_level, from: $at, status: attr_rel_status }]->(a)
            }
            RETURN 1 AS done
        }
        RETURN 1 AS done
    }
    WITH n, node_diff_map
    CALL (n, node_diff_map) {
        UNWIND node_diff_map.relationships AS relationship_diff_map
        // ------------------------------
        // handle updates for relationships under this node
        // ------------------------------
        CALL (n, relationship_diff_map) {
            WITH
                relationship_diff_map.peer_id AS rel_peer_id, relationship_diff_map.name AS rel_name,
                CASE
                    WHEN relationship_diff_map.action = "ADDED" THEN "active"
                    WHEN relationship_diff_map.action = "REMOVED" THEN "deleted"
                    ELSE NULL
                END AS related_rel_status,
                CASE
                    WHEN $migrated_kinds_uuids IS NULL THEN NULL
                    WHEN relationship_diff_map.peer_id IN $migrated_kinds_uuids THEN $migrated_kinds_id_map[relationship_diff_map.peer_id]
                    ELSE NULL
                END AS rel_peer_db_id
            // ------------------------------
            // find the correct relationship peer if the peer had its kind/inheritance migrated
            // and there are multiple Nodes with the same UUID
            // ------------------------------
            CALL (rel_peer_id, rel_peer_db_id) {
                MATCH (rel_peer:Node {uuid: rel_peer_id})-[target_is_part_of:IS_PART_OF]->(:Root)
                WHERE (rel_peer_db_id IS NULL OR %(id_func)s(rel_peer) = rel_peer_db_id)
                AND target_is_part_of.branch IN [$source_branch, $target_branch]
                RETURN rel_peer
                ORDER BY target_is_part_of.branch_level DESC, target_is_part_of.from DESC, target_is_part_of.status ASC
                LIMIT 1
            }
            WITH rel_name, related_rel_status, rel_peer
            // ------------------------------
            // determine the directions of each IS_RELATED
            // ------------------------------
            CALL (n, rel_name, rel_peer, related_rel_status) {
                MATCH (n)
                    -[source_r_rel_1:IS_RELATED]
                    -(r:Relationship {name: rel_name})
                    -[source_r_rel_2:IS_RELATED]
                    -(rel_peer)
                WHERE source_r_rel_1.branch IN [$source_branch, $target_branch]
                AND source_r_rel_2.branch IN [$source_branch, $target_branch]
                AND source_r_rel_1.from <= $at AND source_r_rel_1.to IS NULL
                AND source_r_rel_2.from <= $at AND source_r_rel_2.to IS NULL
                WITH r, source_r_rel_1, source_r_rel_2
                ORDER BY source_r_rel_1.branch_level DESC, source_r_rel_2.branch_level DESC, source_r_rel_1.from DESC, source_r_rel_2.from DESC
                LIMIT 1
                RETURN r, CASE
                    WHEN startNode(source_r_rel_1).uuid = n.uuid THEN "r"
                    ELSE "l"
                END AS r1_dir,
                CASE
                    WHEN startNode(source_r_rel_2).uuid = r.uuid THEN "r"
                    ELSE "l"
                END AS r2_dir,
                source_r_rel_1.hierarchy AS r1_hierarchy,
                source_r_rel_2.hierarchy AS r2_hierarchy
            }
            WITH n, r, r1_dir, r2_dir, r1_hierarchy, r2_hierarchy, rel_name, rel_peer, related_rel_status
            CALL (n, rel_name, rel_peer, related_rel_status) {
                OPTIONAL MATCH (n)
                    -[target_r_rel_1:IS_RELATED {branch: $target_branch, status: "active"}]
                    -(:Relationship {name: rel_name})
                    -[target_r_rel_2:IS_RELATED {branch: $target_branch, status: "active"}]
                    -(rel_peer)
                WHERE related_rel_status = "deleted"
                AND target_r_rel_1.from <= $at AND target_r_rel_1.to IS NULL
                AND target_r_rel_2.from <= $at AND target_r_rel_2.to IS NULL
                SET target_r_rel_1.to = $at
                SET target_r_rel_2.to = $at
            }
            WITH n, r, r1_dir, r2_dir, r1_hierarchy, r2_hierarchy, rel_name, rel_peer, related_rel_status
            // ------------------------------
            // conditionally create new IS_RELATED relationships on target_branch, if necessary
            // ------------------------------
            CALL (n, r, r1_dir, r2_dir, r1_hierarchy, r2_hierarchy, rel_name, rel_peer, related_rel_status) {
                OPTIONAL MATCH (n)
                    -[r_rel_1:IS_RELATED {branch: $target_branch, status: related_rel_status}]
                    -(:Relationship {name: rel_name})
                    -[r_rel_2:IS_RELATED {branch: $target_branch, status: related_rel_status}]
                    -(rel_peer)
                WHERE r_rel_1.from <= $at
                AND (r_rel_1.to >= $at OR r_rel_1.to IS NULL)
                AND r_rel_2.from <= $at
                AND (r_rel_2.to >= $at OR r_rel_2.to IS NULL)
                WITH rel_peer, r_rel_1, r_rel_2
                WHERE r_rel_1 IS NULL
                AND r_rel_2 IS NULL
                // ------------------------------
                // create IS_RELATED relationships with directions maintained from source
                // ------------------------------
                CALL (n, r, r1_dir, r1_hierarchy, related_rel_status) {
                    WITH n, r, r1_dir, r1_hierarchy, related_rel_status
                    WHERE r1_dir = "r"
                    CREATE (n)
                        -[:IS_RELATED {branch: $target_branch, branch_level: $branch_level, from: $at, status: related_rel_status, hierarchy: r1_hierarchy}]
                        ->(r)
                }
                CALL (n, r, r1_dir, r1_hierarchy, related_rel_status) {
                    WITH n, r, r1_dir, r1_hierarchy, related_rel_status
                    WHERE r1_dir = "l"
                    CREATE (n)
                        <-[:IS_RELATED {branch: $target_branch, branch_level: $branch_level, from: $at, status: related_rel_status, hierarchy: r1_hierarchy}]
                        -(r)
                }
                CALL (r, rel_peer, r2_dir, r2_hierarchy, related_rel_status) {
                    WITH r, rel_peer, r2_dir, r2_hierarchy, related_rel_status
                    WHERE r2_dir = "r"
                    CREATE (r)
                        -[:IS_RELATED {branch: $target_branch, branch_level: $branch_level, from: $at, status: related_rel_status, hierarchy: r2_hierarchy}]
                        ->(rel_peer)
                }
                CALL (r, rel_peer, r2_dir, r2_hierarchy, related_rel_status) {
                    WITH r, rel_peer, r2_dir, r2_hierarchy, related_rel_status
                    WHERE r2_dir = "l"
                    CREATE (r)
                        <-[:IS_RELATED {branch: $target_branch, branch_level: $branch_level, from: $at, status: related_rel_status, hierarchy: r2_hierarchy}]
                        -(rel_peer)
                }
            }
        }
    }
}
RETURN 1 AS done
        """ % {"id_func": db.get_id_function_name()}
        self.add_to_query(query=query)


class DiffMergePropertiesQuery(Query):
    name = "diff_merge_properties"
    type = QueryType.WRITE
    insert_return = False

    def __init__(
        self,
        property_diff_dicts: dict[str, Any],
        at: Timestamp,
        target_branch: Branch,
        migrated_kinds_id_map: dict[str, str],
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.property_diff_dicts = property_diff_dicts
        self.at = at
        self.target_branch = target_branch
        self.source_branch_name = self.branch.name
        self.migrated_kinds_id_map = migrated_kinds_id_map

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        self.params = {
            "property_diff_dicts": self.property_diff_dicts,
            "at": self.at.to_string(),
            "branch_level": self.target_branch.hierarchy_level,
            "target_branch": self.target_branch.name,
            "source_branch": self.source_branch_name,
            "migrated_kinds_id_map": self.migrated_kinds_id_map,
            "migrated_kinds_uuids": list(self.migrated_kinds_id_map.keys()),
        }
        query = """
UNWIND $property_diff_dicts AS attr_rel_prop_diff
WITH attr_rel_prop_diff, CASE
    WHEN $migrated_kinds_uuids IS NULL THEN NULL
    WHEN attr_rel_prop_diff.node_uuid IN $migrated_kinds_uuids THEN $migrated_kinds_id_map[attr_rel_prop_diff.node_uuid]
    ELSE NULL
END AS node_db_id,
CASE
    WHEN $migrated_kinds_uuids IS NULL THEN NULL
    WHEN attr_rel_prop_diff.peer_uuid IN $migrated_kinds_uuids THEN $migrated_kinds_id_map[attr_rel_prop_diff.peer_uuid]
    ELSE NULL
END AS peer_db_id
CALL (attr_rel_prop_diff, node_db_id, peer_db_id) {
    // ------------------------------
    // find the Attribute node
    // ------------------------------
    CALL (attr_rel_prop_diff, node_db_id) {
        OPTIONAL MATCH (n:Node {uuid: attr_rel_prop_diff.node_uuid})
            -[has_attr:HAS_ATTRIBUTE]
            ->(attr:Attribute {name: attr_rel_prop_diff.attribute_name})
        WHERE attr_rel_prop_diff.attribute_name IS NOT NULL
        AND (node_db_id IS NULL OR %(id_func)s(n) = node_db_id)
        AND has_attr.branch IN [$source_branch, $target_branch]
        RETURN attr
        ORDER BY has_attr.from DESC
        LIMIT 1
    }
    CALL (attr_rel_prop_diff, node_db_id, peer_db_id) {
        OPTIONAL MATCH (n:Node {uuid: attr_rel_prop_diff.node_uuid})
            -[r1:IS_RELATED]
            -(rel:Relationship {name: attr_rel_prop_diff.relationship_id})
            -[r2:IS_RELATED]
            -(rel_peer:Node {uuid: attr_rel_prop_diff.peer_uuid})
        WHERE attr_rel_prop_diff.relationship_id IS NOT NULL
        AND (node_db_id IS NULL OR %(id_func)s(n) = node_db_id)
        AND (peer_db_id IS NULL OR %(id_func)s(rel_peer) = peer_db_id)
        AND r1.branch IN [$source_branch, $target_branch]
        AND r2.branch IN [$source_branch, $target_branch]
        RETURN rel
        ORDER BY r1.branch_level DESC, r2.branch_level DESC, r1.from DESC, r2.from DESC
        LIMIT 1
    }
    WITH attr_rel_prop_diff, COALESCE(attr, rel) AS attr_rel, peer_db_id
    WHERE attr_rel IS NOT NULL
    UNWIND attr_rel_prop_diff.properties AS property_diff
    // ------------------------------
    // handle updates for properties under this attribute/relationship
    // ------------------------------
    CALL (attr_rel, property_diff, peer_db_id) {
        // ------------------------------
        // identify the correct property node to link
        // ------------------------------
        CALL (attr_rel, property_diff, peer_db_id) {
            OPTIONAL MATCH (peer:Node {uuid: property_diff.value})
            WHERE property_diff.property_type IN ["HAS_SOURCE", "HAS_OWNER"]
            AND (peer_db_id IS NULL OR %(id_func)s(peer) = peer_db_id)
            // ------------------------------
            // the serialized diff might not include the values for IS_VISIBLE and IS_PROTECTED in
            // some cases, so we need to figure them out here
            // ------------------------------
            CALL (attr_rel, property_diff) {
                OPTIONAL MATCH (attr_rel)-[r_vis_pro]->(bool:Boolean)
                WHERE property_diff.property_type IN ["IS_VISIBLE", "IS_PROTECTED"]
                AND r_vis_pro.branch IN [$source_branch, $target_branch]
                AND type(r_vis_pro) = property_diff.property_type
                AND (property_diff.value IS NULL OR bool.value = property_diff.value)
                RETURN bool
                ORDER BY r_vis_pro.from DESC
                LIMIT 1
            }
            CALL (attr_rel, property_diff) {
                // ------------------------------
                // get the latest linked AttributeValue on the source b/c there could be multiple
                // with different is_default values
                // ------------------------------
                OPTIONAL MATCH (attr_rel)-[r_attr_val:HAS_VALUE]->(av:AttributeValue)
                WHERE property_diff.property_type = "HAS_VALUE"
                AND (
                    av.value = property_diff.value
                    OR toLower(toString(av.value)) = toLower(toString(property_diff.value))
                )
                AND r_attr_val.branch IN [$source_branch, $target_branch]
                RETURN av
                ORDER BY r_attr_val.from DESC
                LIMIT 1
            }
            RETURN COALESCE (peer, bool, av) AS prop_node
        }
        WITH attr_rel,property_diff.property_type AS prop_type, prop_node, CASE
            WHEN property_diff.action = "ADDED" THEN "active"
            WHEN property_diff.action = "REMOVED" THEN "deleted"
            ELSE NULL
        END as prop_rel_status
        // ------------------------------
        // set property edge.to, optionally, on target branch
        // ------------------------------
        CALL (attr_rel, prop_rel_status, prop_type) {
            OPTIONAL MATCH (attr_rel)
                -[target_r_prop {branch: $target_branch}]
                ->()
            WHERE type(target_r_prop) = prop_type
            AND target_r_prop.from < $at AND target_r_prop.to IS NULL
            SET target_r_prop.to = $at
        }
        // ------------------------------
        // check for existing edge on target_branch
        // ------------------------------
        CALL (attr_rel, prop_rel_status, prop_type, prop_node) {
            OPTIONAL MATCH (attr_rel)-[r_prop {branch: $target_branch}]->(prop_node)
            WHERE type(r_prop) = prop_type
            AND r_prop.status = prop_rel_status
            AND r_prop.from <= $at
            AND (r_prop.to > $at OR r_prop.to IS NULL)
            RETURN r_prop
        }
        WITH attr_rel,prop_rel_status, prop_type, prop_node, r_prop
        WHERE r_prop IS NULL
        // ------------------------------
        // create new edge to prop_node on target_branch, if necessary
        // one subquery per possible edge type b/c edge type cannot be a variable
        // ------------------------------
        CALL (attr_rel, prop_rel_status, prop_type, prop_node) {
            WITH attr_rel, prop_rel_status, prop_type, prop_node
            WHERE prop_type = "HAS_VALUE"
            CREATE (attr_rel)-[:HAS_VALUE { branch: $target_branch, branch_level: $branch_level, from: $at, status: prop_rel_status }]->(prop_node)
        }
        CALL (attr_rel, prop_rel_status, prop_type, prop_node) {
            WITH attr_rel, prop_rel_status, prop_type, prop_node
            WHERE prop_type = "HAS_SOURCE"
            CREATE (attr_rel)-[:HAS_SOURCE { branch: $target_branch, branch_level: $branch_level, from: $at, status: prop_rel_status }]->(prop_node)
        }
        CALL (attr_rel, prop_rel_status, prop_type, prop_node) {
            WITH attr_rel, prop_rel_status, prop_type, prop_node
            WHERE prop_type = "HAS_OWNER"
            CREATE (attr_rel)-[:HAS_OWNER { branch: $target_branch, branch_level: $branch_level, from: $at, status: prop_rel_status }]->(prop_node)
        }
        CALL (attr_rel, prop_rel_status, prop_type, prop_node) {
            WITH attr_rel, prop_rel_status, prop_type, prop_node
            WHERE prop_type = "IS_VISIBLE"
            CREATE (attr_rel)-[:IS_VISIBLE { branch: $target_branch, branch_level: $branch_level, from: $at, status: prop_rel_status }]->(prop_node)
        }
        CALL (attr_rel, prop_rel_status, prop_type, prop_node) {
            WITH attr_rel, prop_rel_status, prop_type, prop_node
            WHERE prop_type = "IS_PROTECTED"
            CREATE (attr_rel)-[:IS_PROTECTED { branch: $target_branch, branch_level: $branch_level, from: $at, status: prop_rel_status }]->(prop_node)
        }
    }
}
        """ % {"id_func": db.get_id_function_name()}
        self.add_to_query(query=query)


class DiffMergeMigratedKindsQuery(Query):
    name = "diff_merge_migrated_kinds"
    type = QueryType.WRITE
    insert_return = False

    def __init__(
        self,
        migrated_uuids: list[str],
        at: Timestamp,
        target_branch: Branch,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.migrated_uuids = migrated_uuids
        self.at = at
        self.target_branch = target_branch
        self.source_branch_name = self.branch.name

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        self.params = {
            "migrated_uuids": self.migrated_uuids,
            "at": self.at.to_string(),
            "branch_level": self.target_branch.hierarchy_level,
            "target_branch": self.target_branch.name,
            "source_branch": self.source_branch_name,
        }
        query = """
MATCH (n:Node)
WHERE n.uuid IN $migrated_uuids
CALL (n) {
    // --------------
    // for each migrated node (created or deleted), find its latest edges on the source branch,
    // check if they exist on the target, create them if not
    // --------------
    MATCH (n)-[]-(peer)
    WITH DISTINCT n, peer
    CALL (n, peer) {
        // --------------
        // get the latest outbound edge for each type between n and peer
        // --------------
        MATCH (n)-[e {branch: $source_branch}]->(peer)
        WHERE e.from <= $at AND e.to IS NULL
        WITH e, type(e) AS edge_type
        ORDER BY edge_type, e.from DESC
        WITH edge_type, head(collect(e)) AS latest_source_edge
        RETURN edge_type, latest_source_edge
    }
    CALL (n, peer, edge_type) {
        // --------------
        // for each n, peer, edge_type, get the latest edge on target
        // --------------
        OPTIONAL MATCH (n)-[e {branch: $target_branch}]->(peer)
        WHERE type(e) = edge_type AND e.from <= $at
        RETURN e AS latest_target_edge
        ORDER BY e.from DESC
        LIMIT 1
    }
    // --------------
    // ignore edges of this type that already have the correct status on the target branch
    // --------------
    WITH n, peer, edge_type, latest_source_edge, latest_target_edge
    WHERE (latest_target_edge IS NULL AND latest_source_edge.status = "active")
    OR latest_source_edge.status <> latest_target_edge.status
    CALL (latest_source_edge, latest_target_edge) {
        // --------------
        // set the to time on active target branch edges that we are setting to deleted
        // --------------
        WITH latest_target_edge WHERE latest_target_edge IS NOT NULL
        AND latest_source_edge.status = "deleted"
        AND latest_target_edge.status = "active"
        AND latest_target_edge.to IS NULL
        SET latest_target_edge.to = $at
    }
    // --------------
    // create the outbound edges on the target branch, one subquery per possible type
    // --------------
    CALL (n, latest_source_edge, peer, edge_type) {
        WITH edge_type WHERE edge_type = "IS_PART_OF"
        CREATE (n)-[new_edge:IS_PART_OF]->(peer)
        SET new_edge = properties(latest_source_edge)
        SET new_edge.from = $at
        SET new_edge.branch_level = $branch_level
        SET new_edge.branch = $target_branch
    }
    CALL (n, latest_source_edge, peer, edge_type) {
        WITH edge_type
        WHERE edge_type = "IS_RELATED"
        CREATE (n)-[new_edge:IS_RELATED]->(peer)
        SET new_edge = properties(latest_source_edge)
        SET new_edge.from = $at
        SET new_edge.branch_level = $branch_level
        SET new_edge.branch = $target_branch
    }
    CALL (n, latest_source_edge, peer, edge_type) {
        WITH edge_type
        WHERE edge_type = "HAS_ATTRIBUTE"
        CREATE (n)-[new_edge:HAS_ATTRIBUTE]->(peer)
        SET new_edge = properties(latest_source_edge)
        SET new_edge.from = $at
        SET new_edge.branch_level = $branch_level
        SET new_edge.branch = $target_branch
    }
    // --------------
    // do all of this again for inbound edges
    // --------------
    WITH DISTINCT n, peer
    CALL (n, peer) {
        // --------------
        // get the latest inbound edge for each type between n and peer
        // --------------
        MATCH (n)<-[e {branch: $source_branch}]-(peer)
        WHERE e.from <= $at AND e.to IS NULL
        WITH e, type(e) AS edge_type
        ORDER BY edge_type, e.from DESC
        WITH edge_type, head(collect(e)) AS latest_source_edge
        RETURN edge_type, latest_source_edge
    }
    CALL (n, peer, edge_type) {
        // --------------
        // for each n, peer, edge_type, get the latest edge on target
        // --------------
        OPTIONAL MATCH (n)<-[e {branch: $target_branch}]-(peer)
        WHERE type(e) = edge_type AND e.from <= $at
        RETURN e AS latest_target_edge
        ORDER BY e.from DESC
        LIMIT 1
    }
    // --------------
    // ignore edges of this type that already have the correct status on the target branch
    // --------------
    WITH n, peer, edge_type, latest_source_edge, latest_target_edge
    WHERE latest_target_edge IS NULL OR latest_source_edge.status <> latest_target_edge.status
    CALL (latest_source_edge, latest_target_edge) {
        // --------------
        // set the to time on active target branch edges that we are setting to deleted
        // --------------
        WITH latest_target_edge
        WHERE latest_target_edge IS NOT NULL
        AND latest_source_edge.status = "deleted"
        AND latest_target_edge.status = "active"
        AND latest_target_edge.to IS NULL
        SET latest_target_edge.to = $at
    }
    // --------------
    // create the outbound edges on the target branch, one subquery per possible type
    // --------------
    CALL (n, latest_source_edge, peer, edge_type) {
        WITH edge_type
        WHERE edge_type = "IS_RELATED"
        CREATE (n)<-[new_edge:IS_RELATED]-(peer)
        SET new_edge = properties(latest_source_edge)
        SET new_edge.from = $at
        SET new_edge.branch_level = $branch_level
        SET new_edge.branch = $target_branch
    }
    CALL (n, latest_source_edge, peer, edge_type) {
        WITH edge_type
        WHERE edge_type = "HAS_OWNER"
        CREATE (n)<-[new_edge:HAS_OWNER]-(peer)
        SET new_edge = properties(latest_source_edge)
        SET new_edge.from = $at
        SET new_edge.branch_level = $branch_level
        SET new_edge.branch = $target_branch
    }
    CALL (n, latest_source_edge, peer, edge_type) {
        WITH edge_type
        WHERE edge_type = "HAS_SOURCE"
        CREATE (n)<-[new_edge:HAS_SOURCE]-(peer)
        SET new_edge = properties(latest_source_edge)
        SET new_edge.from = $at
        SET new_edge.branch_level = $branch_level
        SET new_edge.branch = $target_branch
    }
}
        """
        self.add_to_query(query)


class DiffMergeRollbackQuery(Query):
    name = "diff_merge_rollback"
    type = QueryType.WRITE
    insert_return = False

    def __init__(
        self,
        at: Timestamp,
        target_branch: Branch,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.at = at
        self.target_branch = target_branch
        self.source_branch_name = self.branch.name

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        self.params = {
            "at": self.at.to_string(),
            "target_branch": self.target_branch.name,
            "source_branch": self.source_branch_name,
        }
        query = """
        // ---------------------------
        // reset to times on target branch
        // ---------------------------
        CALL () {
            OPTIONAL MATCH ()-[r_to {to: $at, branch: $target_branch}]-()
            SET r_to.to = NULL
        }
        // ---------------------------
        // reset from times on target branch
        // ---------------------------
        CALL () {
            OPTIONAL MATCH ()-[r_from {from: $at, branch: $target_branch}]-()
            DELETE r_from
        }
        """
        self.add_to_query(query=query)



from __future__ import annotations

import ipaddress
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from rich.progress import Progress

from infrahub.core.branch.models import Branch
from infrahub.core.constants import InfrahubKind
from infrahub.core.initialization import initialization
from infrahub.core.ipam.reconciler import IpamReconciler
from infrahub.core.migrations.shared import MigrationResult, get_migration_console
from infrahub.core.query import Query, QueryType
from infrahub.lock import initialize_lock
from infrahub.log import get_logger

from ..shared import ArbitraryMigration

if TYPE_CHECKING:
    from infrahub.core.ipam.constants import AllIPTypes
    from infrahub.database import InfrahubDatabase

log = get_logger()


@dataclass
class IpNodeDetails:
    branch: str
    ip_value: AllIPTypes
    namespace: str
    node_uuid: str


class FindNodesToReconcileQuery(Query):
    name = "find_nodes_to_reconcile_query"
    type = QueryType.READ
    insert_return = False

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        self.params.update(
            {
                "prefix_kind": InfrahubKind.IPPREFIX,
                "address_kind": InfrahubKind.IPADDRESS,
                "prefix_attribute_name": "prefix",
                "address_attribute_name": "address",
                "attr_names": ["prefix", "address"],
                "namespace_relationship_names": ["ip_namespace__ip_prefix", "ip_namespace__ip_address"],
            }
        )
        query = """
MATCH (root:Root)
LIMIT 1
WITH root.default_branch AS default_branch
// ------------------
// find prefixes and addresses that exist on the default branch
// ------------------
MATCH (ip_node:%(prefix_kind)s|%(address_kind)s)-[e:IS_PART_OF {branch: default_branch, status: "active"}]->(:Root)
WHERE e.to IS NULL
AND NOT exists((ip_node)-[:IS_PART_OF {branch: default_branch, status: "deleted"}]->(:Root))
// ------------------
// check if this node has its prefix/address updated on a branch (or branches)
// ------------------
CALL (ip_node, default_branch) {
    OPTIONAL MATCH (ip_node)-[:HAS_ATTRIBUTE {branch: default_branch, status: "active"}]->
        (attr:Attribute)-[has_value_e:HAS_VALUE {status: "active"}]->(attr_val)
    WHERE attr.name IN $attr_names
    AND has_value_e.branch_level = 2
    AND (
        ($prefix_kind IN labels(ip_node) AND attr.name = $prefix_attribute_name)
        OR ($address_kind IN labels(ip_node) AND attr.name = $address_attribute_name)
    )
    RETURN collect(DISTINCT has_value_e.branch) AS attr_update_branches
}
CALL (ip_node, default_branch) {
    OPTIONAL MATCH (ip_node)-[:IS_RELATED {status: "active"}]-
        (rel:Relationship)-[e2:IS_RELATED {status: "active"}]-(peer:Node)
    WHERE rel.name IN $namespace_relationship_names
    AND e2.branch_level = 2
    RETURN collect(DISTINCT e2.branch) AS namespace_update_branches
}
// ------------------
// filter to only those prefixes/addresses with an update that we care about on a branch
// ------------------
WITH DISTINCT ip_node, default_branch, attr_update_branches, namespace_update_branches
WHERE size(attr_update_branches) > 0 OR size(namespace_update_branches) > 0
// ------------------
// deduplicate branch lists and return one row for each branch
// ------------------
CALL (attr_update_branches, namespace_update_branches) {
    UNWIND attr_update_branches + namespace_update_branches AS branch
    RETURN DISTINCT branch
}
WITH default_branch, branch, ip_node
// ------------------
// confirm node is still active on this branch
// ------------------
MATCH (ip_node)
WHERE NOT exists((ip_node)-[:IS_PART_OF {branch: branch, status: "deleted"}]->(:Root))
// ------------------
// get branched_from time for each branch
// ------------------
CALL (branch) {
    MATCH (b:Branch {name: branch})
    RETURN b.branched_from AS branched_from
}
// ------------------
// get latest namespace on this branch
// ------------------
CALL (default_branch, branch, branched_from, ip_node) {
    MATCH (ip_node)-[e1:IS_RELATED]-(rel:Relationship)-[e2:IS_RELATED]-(peer:%(namespace_kind)s)
    WHERE rel.name IN $namespace_relationship_names
    AND (
        e1.branch = branch
        OR (e1.branch = default_branch AND e1.from < branched_from)
    )
    AND e1.to IS NULL
    AND (
        e2.branch = branch
        OR (e2.branch = default_branch AND e2.from < branched_from)
    )
    AND e2.to IS NULL
    WITH peer, e1.status = "active" AND e2.status = "active" AS is_active
    ORDER BY e2.branch_level DESC, e2.from DESC, e2.status ASC, e1.branch_level DESC, e1.from DESC, e1.status ASC
    LIMIT 1
    WITH peer
    WHERE is_active = TRUE
    RETURN peer.uuid AS namespace_uuid
}
// ------------------
// get latest prefix/address value on this branch
// ------------------
CALL (default_branch, branch, branched_from, ip_node) {
    MATCH (ip_node)-[e1:HAS_ATTRIBUTE]->(attr:Attribute)-[e2:HAS_VALUE]->(av)
    WHERE attr.name IN $attr_names
    AND (
        ($prefix_kind IN labels(ip_node) AND attr.name = $prefix_attribute_name)
        OR ($address_kind IN labels(ip_node) AND attr.name = $address_attribute_name)
    )
    AND (
        e1.branch = branch
        OR (e1.branch = default_branch AND e1.from < branched_from)
    )
    AND e1.to IS NULL
    AND (
        e2.branch = branch
        OR (e2.branch = default_branch AND e2.from < branched_from)
    )
    AND e2.to IS NULL
    WITH attr, av, e1.status = "active" AND e2.status = "active" AS is_active
    ORDER BY e2.branch_level DESC, e2.from DESC, e2.status ASC, e1.branch_level DESC, e1.from DESC, e1.status ASC
    LIMIT 1
    WITH attr, av
    WHERE is_active = TRUE
    RETURN CASE
        WHEN attr.name = $prefix_attribute_name THEN av.value
        ELSE NULL
    END AS prefix_value,
    CASE
        WHEN attr.name = $address_attribute_name THEN av.value
        ELSE NULL
    END AS address_value
}
RETURN branch, namespace_uuid, ip_node.uuid AS node_uuid, prefix_value, address_value
        """ % {
            "prefix_kind": InfrahubKind.IPPREFIX,
            "address_kind": InfrahubKind.IPADDRESS,
            "namespace_kind": InfrahubKind.IPNAMESPACE,
        }
        self.add_to_query(query)
        self.return_labels = ["branch", "namespace_uuid", "node_uuid", "prefix_value", "address_value"]

    def get_nodes_to_reconcile(self) -> list[IpNodeDetails]:
        ip_node_details = []
        for result in self.get_results():
            prefix_value = result.get_as_str("prefix_value")
            address_value = result.get_as_str("address_value")
            if prefix_value:
                ip_value: AllIPTypes = ipaddress.ip_network(address=prefix_value)
            elif address_value:
                ip_value = ipaddress.ip_interface(address=address_value)
            else:
                continue
            ip_node_details.append(
                IpNodeDetails(
                    branch=result.get_as_type("branch", str),
                    ip_value=ip_value,
                    namespace=result.get_as_type("namespace_uuid", str),
                    node_uuid=result.get_as_type("node_uuid", str),
                )
            )
        return ip_node_details


class DeleteSelfParentRelationshipsQuery(Query):
    name = "delete_self_parent_relationships"
    type = QueryType.WRITE
    insert_return = False

    def __init__(self, uuids_to_check: list[str], **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.uuids_to_check = uuids_to_check

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        self.params["uuids_to_check"] = self.uuids_to_check
        query = """
MATCH (n:Node)-[e1:IS_RELATED]->(:Relationship {name: "parent__child"})-[e2:IS_RELATED {branch: e1.branch, status: e1.status, from: e1.from}]->(n)
WHERE n.uuid IN $uuids_to_check
DELETE e1, e2
        """
        self.add_to_query(query)


class Migration039(ArbitraryMigration):
    """
    Identify all IP prefixes/addresses that have been updated on a branch and reconcile them on that branch
    If any of the identified IP prefixes/addresses are their own parent/child, delete those illegal edges before reconciling.
    """

    name: str = "039_ipam_reconcile_updated"
    minimum_version: int = 38

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._reconcilers_by_branch: dict[str, IpamReconciler] = {}

    async def _get_reconciler(self, db: InfrahubDatabase, branch_name: str) -> IpamReconciler:
        if branch_name not in self._reconcilers_by_branch:
            branch = await Branch.get_by_name(db=db, name=branch_name)
            self._reconcilers_by_branch[branch_name] = IpamReconciler(db=db, branch=branch)
        return self._reconcilers_by_branch[branch_name]

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:  # noqa: ARG002
        return MigrationResult()

    async def execute(self, db: InfrahubDatabase) -> MigrationResult:
        console = get_migration_console()
        result = MigrationResult()
        # load schemas from database into registry
        initialize_lock()
        await initialization(db=db)

        console.print("Identifying IP prefixes and addresses to reconcile...", end="")
        find_nodes_query = await FindNodesToReconcileQuery.init(db=db)
        await find_nodes_query.execute(db=db)
        console.print("done")

        # we need to delete the self-parent relationships before reconciling b/c the
        # reconciler cannot correctly handle a prefix that is its own parent
        ip_node_details_list = find_nodes_query.get_nodes_to_reconcile()
        uuids_to_check = {ip_node_details.node_uuid for ip_node_details in ip_node_details_list}
        console.log(f"{len(ip_node_details_list)} IP prefixes or addresses will be reconciled.")

        console.print("Deleting self-parent relationships prior to reconciliation...", end="")
        delete_self_parent_relationships_query = await DeleteSelfParentRelationshipsQuery.init(
            db=db, uuids_to_check=list(uuids_to_check)
        )
        await delete_self_parent_relationships_query.execute(db=db)
        console.print("done")

        console.log("Reconciling IP prefixes and addresses across branches...")
        with Progress(console=console) as progress:
            reconcile_task = progress.add_task("Reconciling IP prefixes/addresses...", total=len(ip_node_details_list))

            for ip_node_details in ip_node_details_list:
                reconciler = await self._get_reconciler(db=db, branch_name=ip_node_details.branch)
                await reconciler.reconcile(
                    ip_value=ip_node_details.ip_value,
                    namespace=ip_node_details.namespace,
                    node_uuid=ip_node_details.node_uuid,
                )
                progress.update(reconcile_task, advance=1)

        console.log("IP prefix and address reconciliation complete.")

        return result



from __future__ import annotations

from typing import Any, Sequence

from ..query import AttributeMigrationQuery
from ..query.attribute_remove import AttributeRemoveQuery
from ..shared import AttributeSchemaMigration


class NodeAttributeRemoveMigrationQuery01(AttributeMigrationQuery, AttributeRemoveQuery):
    name = "migration_node_attribute_remove_01"
    insert_return: bool = False

    def __init__(
        self,
        migration: AttributeSchemaMigration,
        **kwargs: Any,
    ):
        super().__init__(
            migration=migration,
            attribute_name=migration.previous_attribute_schema.name,
            node_kinds=[migration.new_schema.kind],
            **kwargs,
        )


class NodeAttributeRemoveMigration(AttributeSchemaMigration):
    name: str = "node.attribute.remove"
    queries: Sequence[type[AttributeMigrationQuery]] = [NodeAttributeRemoveMigrationQuery01]  # type: ignore[assignment]



# Generated by "invoke backend.generate", do not edit directly

from __future__ import annotations

from pydantic import Field

from infrahub.core.constants import BranchSupportType, HashableModelState
from infrahub.core.models import HashableModel
from infrahub.core.schema.attribute_schema import AttributeSchema  # noqa: TC001
from infrahub.core.schema.relationship_schema import RelationshipSchema  # noqa: TC001


class GeneratedBaseNodeSchema(HashableModel):
    id: str | None = Field(
        default=None, description="The ID of the node", json_schema_extra={"update": "not_applicable"}
    )
    name: str = Field(
        ...,
        description="Node name, must be unique within a namespace and must start with an uppercase letter.",
        pattern="^[A-Z][a-zA-Z0-9]+$",
        min_length=2,
        max_length=32,
        json_schema_extra={"update": "migration_required"},
    )
    namespace: str = Field(
        ...,
        description="Node Namespace, Namespaces are used to organize models into logical groups and to prevent name collisions.",
        pattern="^[A-Z][a-z0-9]+$",
        min_length=3,
        max_length=32,
        json_schema_extra={"update": "migration_required"},
    )
    description: str | None = Field(
        default=None,
        description="Short description of the model, will be visible in the frontend.",
        max_length=128,
        json_schema_extra={"update": "allowed"},
    )
    label: str | None = Field(
        default=None,
        description="Human friendly representation of the name/kind",
        max_length=64,
        json_schema_extra={"update": "allowed"},
    )
    branch: BranchSupportType = Field(
        default=BranchSupportType.AWARE,
        description="Type of branch support for the model.",
        json_schema_extra={"update": "not_supported"},
    )
    default_filter: str | None = Field(
        default=None,
        description="Default filter used to search for a node in addition to its ID. (deprecated: please use human_friendly_id instead)",
        pattern=r"^[a-z0-9\_]*$",
        json_schema_extra={"update": "allowed"},
    )
    human_friendly_id: list[str] | None = Field(
        default=None,
        description="Human friendly and unique identifier for the object.",
        json_schema_extra={"update": "allowed"},
    )
    display_label: str | None = Field(
        default=None,
        description="Attribute or Jinja2 template to use to generate the display label",
        json_schema_extra={"update": "allowed"},
    )
    display_labels: list[str] | None = Field(
        default=None,
        description="List of attributes to use to generate the display label (deprecated)",
        json_schema_extra={"update": "allowed"},
    )
    include_in_menu: bool | None = Field(
        default=None,
        description="Defines if objects of this kind should be included in the menu.",
        json_schema_extra={"update": "allowed"},
    )
    menu_placement: str | None = Field(
        default=None,
        description="Defines where in the menu this object should be placed.",
        json_schema_extra={"update": "allowed"},
    )
    icon: str | None = Field(
        default=None,
        description="Defines the icon to use in the menu. Must be a valid value from the MDI library https://icon-sets.iconify.design/mdi/",
        json_schema_extra={"update": "allowed"},
    )
    order_by: list[str] | None = Field(
        default=None,
        description="List of attributes to use to order the results by default",
        json_schema_extra={"update": "allowed"},
    )
    uniqueness_constraints: list[list[str]] | None = Field(
        default=None,
        description="List of multi-element uniqueness constraints that can combine relationships and attributes",
        json_schema_extra={"update": "validate_constraint"},
    )
    documentation: str | None = Field(
        default=None,
        description="Link to a documentation associated with this object, can be internal or external.",
        json_schema_extra={"update": "allowed"},
    )
    state: HashableModelState = Field(
        default=HashableModelState.PRESENT,
        description="Expected state of the node/generic after loading the schema",
        json_schema_extra={"update": "not_applicable"},
    )
    attributes: list[AttributeSchema] = Field(
        default_factory=list, description="Node attributes", json_schema_extra={"update": "not_applicable"}
    )
    relationships: list[RelationshipSchema] = Field(
        default_factory=list, description="Node Relationships", json_schema_extra={"update": "not_applicable"}
    )



from typing import TYPE_CHECKING, Any, Self

from graphene import Boolean, InputObjectType, Mutation, String
from graphene.types.generic import GenericScalar
from graphql import GraphQLResolveInfo

from infrahub.core import registry
from infrahub.core.constants.infrahubkind import READONLYREPOSITORY, REPOSITORY
from infrahub.core.convert_object_type.object_conversion import ConversionFieldInput, convert_and_validate_object_type
from infrahub.core.convert_object_type.repository_conversion import convert_repository_type
from infrahub.core.convert_object_type.schema_mapping import get_schema_mapping
from infrahub.core.manager import NodeManager
from infrahub.exceptions import ValidationError
from infrahub.repositories.create_repository import RepositoryFinalizer

if TYPE_CHECKING:
    from infrahub.core.attribute import BaseAttribute
    from infrahub.graphql.initialization import GraphqlContext


class ConvertObjectTypeInput(InputObjectType):
    node_id = String(required=True)
    target_kind = String(required=True)
    fields_mapping = GenericScalar(required=True)  # keys are destination attributes/relationships names.


class ConvertObjectType(Mutation):
    class Arguments:
        data = ConvertObjectTypeInput(required=True)

    ok = Boolean()
    node = GenericScalar()

    @classmethod
    async def mutate(
        cls,
        root: dict,  # noqa: ARG003
        info: GraphQLResolveInfo,
        data: ConvertObjectTypeInput,
    ) -> Self:
        """Convert an input node to a given compatible kind."""

        graphql_context: GraphqlContext = info.context

        node_to_convert = await NodeManager.get_one(
            id=str(data.node_id), db=graphql_context.db, branch=graphql_context.branch
        )

        source_schema = registry.get_node_schema(name=node_to_convert.get_kind(), branch=graphql_context.branch)
        target_schema = registry.get_node_schema(name=str(data.target_kind), branch=graphql_context.branch)

        fields_mapping: dict[str, ConversionFieldInput] = {}
        if not isinstance(data.fields_mapping, dict):
            raise ValidationError(
                input_value=f"Expected `fields_mapping` to be a `dict`, got {type(data.fields_mapping)}"
            )

        for field_name, input_for_dest_field_str in data.fields_mapping.items():
            fields_mapping[field_name] = ConversionFieldInput(**input_for_dest_field_str)

        node_to_convert = await NodeManager.get_one(
            id=str(data.node_id), db=graphql_context.db, branch=graphql_context.branch
        )
        for attribute_name in source_schema.attribute_names:
            attribute: BaseAttribute = getattr(node_to_convert, attribute_name)
            if attribute.is_from_profile:
                raise ValidationError(
                    input_value=f"The attribute '{attribute_name}' is from a profile, converting objects that use profiles is not yet supported."
                )

        # Complete fields mapping with auto-mapping.
        mapping = get_schema_mapping(source_schema=source_schema, target_schema=target_schema)
        for field_name, mapping_value in mapping.items():
            if mapping_value.source_field_name is not None and field_name not in fields_mapping:
                fields_mapping[field_name] = ConversionFieldInput(source_field=mapping_value.source_field_name)

        if target_schema.kind in [REPOSITORY, READONLYREPOSITORY]:
            new_node = await convert_repository_type(
                repository=node_to_convert,
                target_schema=target_schema,
                mapping=fields_mapping,
                branch=graphql_context.branch,
                db=graphql_context.db,
                repository_post_creator=RepositoryFinalizer(
                    account_session=graphql_context.active_account_session,
                    services=graphql_context.active_service,
                    context=graphql_context.get_context(),
                ),
            )
        else:
            new_node = await convert_and_validate_object_type(
                node=node_to_convert,
                target_schema=target_schema,
                mapping=fields_mapping,
                branch=graphql_context.branch,
                db=graphql_context.db,
            )

        dict_node = await new_node.to_graphql(db=graphql_context.db, fields={})
        result: dict[str, Any] = {"ok": True, "node": dict_node}

        return cls(**result)



from __future__ import annotations

import hashlib
from copy import deepcopy
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from infrahub.core.schema import SchemaAttributePath


@dataclass
class TemplateLabel:
    template: str
    attributes: set[str] = field(default_factory=set)
    relationships: set[str] = field(default_factory=set)
    relationship_fields: dict[str, set[str]] = field(default_factory=dict)
    filter_key: str = "ids"

    @property
    def fields(self) -> list[str]:
        return sorted(list(self.attributes) + list(self.relationships))

    @property
    def has_related_components(self) -> bool:
        """Indicate if the associated template use variables from relationships"""
        return len(self.relationships) > 0

    def get_hash(self) -> str:
        return hashlib.md5(self.template.encode(), usedforsecurity=False).hexdigest()


@dataclass
class RelationshipIdentifier:
    kind: str
    filter_key: str
    template: str

    def __hash__(self) -> int:
        return hash(f"{self.kind}::{self.filter_key}::{self.template}")


@dataclass
class RelationshipTriggers:
    attributes: dict[str, set[RelationshipIdentifier]] = field(default_factory=dict)


class DisplayLabels:
    def __init__(
        self,
        template_based_display_labels: dict[str, TemplateLabel] | None = None,
        template_relationship_triggers: dict[str, RelationshipTriggers] | None = None,
    ) -> None:
        self._template_based_display_labels: dict[str, TemplateLabel] = template_based_display_labels or {}
        self._template_relationship_triggers: dict[str, RelationshipTriggers] = template_relationship_triggers or {}

    def duplicate(self) -> DisplayLabels:
        """Clone the current object."""
        return self.__class__(
            template_based_display_labels=deepcopy(self._template_based_display_labels),
            template_relationship_triggers=deepcopy(self._template_relationship_triggers),
        )

    def register_attribute_based_display_label(self, kind: str, attribute_name: str) -> None:
        """Register nodes where the display label consists of a single defined attribute name."""
        self._template_based_display_labels[kind] = TemplateLabel(
            template=f"{{{{ {attribute_name}__value }}}}", attributes={attribute_name}
        )

    def register_template_schema_path(self, kind: str, schema_path: SchemaAttributePath, template: str) -> None:
        """Register Jinja2 template based display labels using the schema path of each impacted variable in the node."""

        if kind not in self._template_based_display_labels:
            self._template_based_display_labels[kind] = TemplateLabel(template=template)

        if schema_path.is_type_attribute:
            self._template_based_display_labels[kind].attributes.add(schema_path.active_attribute_schema.name)
        elif schema_path.is_type_relationship and schema_path.related_schema:
            self._template_based_display_labels[kind].relationships.add(schema_path.active_relationship_schema.name)
            if (
                schema_path.active_relationship_schema.name
                not in self._template_based_display_labels[kind].relationship_fields
            ):
                self._template_based_display_labels[kind].relationship_fields[
                    schema_path.active_relationship_schema.name
                ] = set()
            self._template_based_display_labels[kind].relationship_fields[
                schema_path.active_relationship_schema.name
            ].add(schema_path.active_attribute_schema.name)

            if schema_path.related_schema.kind not in self._template_relationship_triggers:
                self._template_relationship_triggers[schema_path.related_schema.kind] = RelationshipTriggers()
            if (
                schema_path.active_attribute_schema.name
                not in self._template_relationship_triggers[schema_path.related_schema.kind].attributes
            ):
                self._template_relationship_triggers[schema_path.related_schema.kind].attributes[
                    schema_path.active_attribute_schema.name
                ] = set()
            self._template_relationship_triggers[schema_path.related_schema.kind].attributes[
                schema_path.active_attribute_schema.name
            ].add(
                RelationshipIdentifier(
                    kind=kind, filter_key=f"{schema_path.active_relationship_schema.name}__ids", template=template
                )
            )

    def targets_node(self, kind: str) -> bool:
        """Indicates if there is a display_label defined for the targeted node"""
        return kind in self._template_based_display_labels

    def get_template_node(self, kind: str) -> TemplateLabel:
        """Return node kinds together with their template definitions."""
        return self._template_based_display_labels[kind]

    def get_template_nodes(self) -> dict[str, TemplateLabel]:
        """Return node kinds together with their template definitions."""
        return self._template_based_display_labels

    def get_related_trigger_nodes(self) -> dict[str, RelationshipTriggers]:
        """Return node kinds that other nodes use within their templates for display_labels."""
        return self._template_relationship_triggers

    def get_related_template(self, related_kind: str, target_kind: str) -> TemplateLabel:
        relationship_trigger = self._template_relationship_triggers[related_kind]
        for applicable_kinds in relationship_trigger.attributes.values():
            for relationship_identifier in applicable_kinds:
                if target_kind == relationship_identifier.kind:
                    template_label = self.get_template_node(kind=target_kind)
                    template_label.filter_key = relationship_identifier.filter_key
                    return template_label

        raise ValueError(
            f"Unable to find registered template for {target_kind} registered on related node {related_kind}"
        )



from infrahub.actions.triggers import TRIGGER_ACTION_RULE_UPDATE
from infrahub.branch.triggers import TRIGGER_BRANCH_MERGED
from infrahub.computed_attribute.triggers import (
    TRIGGER_COMPUTED_ATTRIBUTE_ALL_SCHEMA,
    TRIGGER_COMPUTED_ATTRIBUTE_PYTHON_SETUP_COMMIT,
)
from infrahub.display_labels.triggers import TRIGGER_DISPLAY_LABELS_ALL_SCHEMA
from infrahub.hfid.triggers import TRIGGER_HFID_ALL_SCHEMA
from infrahub.schema.triggers import TRIGGER_SCHEMA_UPDATED
from infrahub.trigger.models import TriggerDefinition
from infrahub.webhook.triggers import TRIGGER_WEBHOOK_DELETE, TRIGGER_WEBHOOK_SETUP_UPDATE

builtin_triggers: list[TriggerDefinition] = [
    TRIGGER_ACTION_RULE_UPDATE,
    TRIGGER_BRANCH_MERGED,
    TRIGGER_COMPUTED_ATTRIBUTE_ALL_SCHEMA,
    TRIGGER_COMPUTED_ATTRIBUTE_PYTHON_SETUP_COMMIT,
    TRIGGER_DISPLAY_LABELS_ALL_SCHEMA,
    TRIGGER_HFID_ALL_SCHEMA,
    TRIGGER_SCHEMA_UPDATED,
    TRIGGER_WEBHOOK_DELETE,
    TRIGGER_WEBHOOK_SETUP_UPDATE,
]



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Sequence

from infrahub.core.migrations.query.attribute_remove import AttributeRemoveQuery
from infrahub.core.schema.generic_schema import GenericSchema
from infrahub.core.schema.node_schema import NodeSchema

from ..query import AttributeMigrationQuery, MigrationBaseQuery
from ..query.attribute_add import AttributeAddQuery
from ..shared import AttributeSchemaMigration, MigrationResult

if TYPE_CHECKING:
    from infrahub.core.branch.models import Branch
    from infrahub.core.schema import MainSchemaTypes
    from infrahub.core.timestamp import Timestamp
    from infrahub.database import InfrahubDatabase


def _get_node_kinds(schema: MainSchemaTypes) -> list[str]:
    if not isinstance(schema, (NodeSchema, GenericSchema)):
        return [schema.kind]
    schema_kinds = [f"Profile{schema.kind}"]
    if isinstance(schema, GenericSchema) and schema.used_by:
        schema_kinds += [f"Profile{kind}" for kind in schema.used_by]
    return schema_kinds


class ProfilesAttributeAddMigrationQuery(AttributeMigrationQuery, AttributeAddQuery):
    name = "migration_profiles_attribute_add"

    def __init__(
        self,
        migration: AttributeSchemaMigration,
        **kwargs: Any,
    ):
        node_kinds = _get_node_kinds(migration.new_schema)
        super().__init__(
            migration=migration,
            node_kinds=node_kinds,
            attribute_name=migration.new_attribute_schema.name,
            attribute_kind=migration.new_attribute_schema.kind,
            branch_support=migration.new_attribute_schema.get_branch(),
            default_value=migration.new_attribute_schema.default_value,
            **kwargs,
        )


class ProfilesAttributeRemoveMigrationQuery(AttributeMigrationQuery, AttributeRemoveQuery):
    name = "migration_profiles_attribute_remove"

    def __init__(
        self,
        migration: AttributeSchemaMigration,
        **kwargs: Any,
    ):
        node_kinds = _get_node_kinds(migration.new_schema)
        super().__init__(
            migration=migration,
            attribute_name=migration.new_attribute_schema.name,
            node_kinds=node_kinds,
            **kwargs,
        )


class AttributeSupportsProfileUpdateMigration(AttributeSchemaMigration):
    name: str = "attribute.supports_profile.update"
    queries: Sequence[type[MigrationBaseQuery]] = []

    async def execute(
        self,
        db: InfrahubDatabase,
        branch: Branch,
        at: Timestamp | str | None = None,
        queries: Sequence[type[MigrationBaseQuery]] | None = None,  # noqa: ARG002
    ) -> MigrationResult:
        if (
            # no change in whether the attribute should be used on profiles
            self.previous_attribute_schema.support_profiles == self.new_attribute_schema.support_profiles
            # the attribute is new, so there cannot be existing profiles to update
            or self.previous_attribute_schema.id is None
        ):
            return MigrationResult()
        profiles_queries: list[type[AttributeMigrationQuery]] = []
        if self.new_attribute_schema.support_profiles:
            profiles_queries.append(ProfilesAttributeAddMigrationQuery)
        if not self.new_attribute_schema.support_profiles:
            profiles_queries.append(ProfilesAttributeRemoveMigrationQuery)

        return await super().execute(db=db, branch=branch, at=at, queries=profiles_queries)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from graphene import Boolean, InputObjectType, List, Mutation, NonNull, String

from infrahub.core.account import ObjectPermission
from infrahub.core.constants import ComputedAttributeKind, PermissionAction, PermissionDecision
from infrahub.core.manager import NodeManager
from infrahub.core.protocols import CoreTransformPython
from infrahub.core.registry import registry
from infrahub.database import retry_db_transaction
from infrahub.events import EventMeta
from infrahub.events.node_action import NodeUpdatedEvent
from infrahub.exceptions import NodeNotFoundError, ProcessingError, ValidationError
from infrahub.graphql.context import apply_external_context
from infrahub.graphql.types.context import ContextInput
from infrahub.log import get_log_data
from infrahub.worker import WORKER_IDENTITY
from infrahub.workers.dependencies import get_workflow
from infrahub.workflows.catalogue import (
    COMPUTED_ATTRIBUTE_PROCESS_JINJA2,
    COMPUTED_ATTRIBUTE_PROCESS_TRANSFORM,
    TRIGGER_UPDATE_JINJA_COMPUTED_ATTRIBUTES,
    TRIGGER_UPDATE_PYTHON_COMPUTED_ATTRIBUTES,
)

if TYPE_CHECKING:
    from graphql import GraphQLResolveInfo

    from infrahub.graphql.initialization import GraphqlContext


class InfrahubComputedAttributeUpdateInput(InputObjectType):
    id = String(required=True)
    kind = String(required=True)
    attribute = String(required=True)
    value = String(required=True)


class UpdateComputedAttribute(Mutation):
    class Arguments:
        data = InfrahubComputedAttributeUpdateInput(required=True)
        context = ContextInput(required=False)

    ok = Boolean()

    @classmethod
    @retry_db_transaction(name="update_computed_attribute")
    async def mutate(
        cls,
        _: dict,
        info: GraphQLResolveInfo,
        data: InfrahubComputedAttributeUpdateInput,
        context: ContextInput | None = None,
    ) -> UpdateComputedAttribute:
        graphql_context: GraphqlContext = info.context
        node_schema = registry.schema.get_node_schema(
            name=str(data.kind), branch=graphql_context.branch.name, duplicate=False
        )
        target_attribute = node_schema.get_attribute(name=str(data.attribute))
        if (
            not target_attribute.computed_attribute
            or target_attribute.computed_attribute.kind == ComputedAttributeKind.USER
        ):
            raise ValidationError(input_value=f"{node_schema.kind}.{target_attribute.name} is not a computed attribute")

        graphql_context.active_permissions.raise_for_permission(
            permission=ObjectPermission(
                namespace=node_schema.namespace,
                name=node_schema.name,
                action=PermissionAction.UPDATE.value,
                decision=PermissionDecision.ALLOW_DEFAULT.value
                if graphql_context.branch.name == registry.default_branch
                else PermissionDecision.ALLOW_OTHER.value,
            )
        )
        await apply_external_context(graphql_context=graphql_context, context_input=context)

        if not (
            target_node := await NodeManager.get_one(
                db=graphql_context.db,
                kind=node_schema.kind,
                id=str(data.id),
                branch=graphql_context.branch,
                fields={target_attribute.name: None},
            )
        ):
            raise NodeNotFoundError(
                node_type="target_node",
                identifier=str(data.id),
                message="The indicated node was not found in the database",
            )

        attribute_field = getattr(target_node, str(data.attribute), None)
        if not attribute_field:
            raise NodeNotFoundError(
                node_type="target_node",
                identifier=str(data.id),
                message="The indicated node does not have the specified attribute_name",
            )
        if attribute_field.value != str(data.value):
            attribute_field.value = str(data.value)
            async with graphql_context.db.start_transaction() as dbt:
                await target_node.save(db=dbt, fields=[str(data.attribute)])

            log_data = get_log_data()
            request_id = log_data.get("request_id", "")

            event = NodeUpdatedEvent(
                kind=node_schema.kind,
                node_id=target_node.get_id(),
                changelog=target_node.node_changelog.model_dump(),
                fields=[str(data.attribute)],
                meta=EventMeta(
                    context=graphql_context.get_context(),
                    initiator_id=WORKER_IDENTITY,
                    request_id=request_id,
                    account_id=graphql_context.active_account_session.account_id,
                    branch=graphql_context.branch,
                ),
            )
            await graphql_context.active_service.event.send(event=event)

        result: dict[str, Any] = {"ok": True}

        return cls(**result)


class InfrahubComputedAttributeRecomputeInput(InputObjectType):
    kind = String(required=True, description="Kind of the node to update")
    attribute = String(required=True, description="Name of the computed attribute that must be recomputed")
    node_ids = List(NonNull(String), description="ID of the nodes for which the attribute must be recomputed")


class RecomputeComputedAttribute(Mutation):
    class Arguments:
        data = InfrahubComputedAttributeRecomputeInput(required=True)
        context = ContextInput(required=False)

    ok = Boolean()

    @classmethod
    @retry_db_transaction(name="update_computed_attribute")
    async def mutate(
        cls,
        _: dict,
        info: GraphQLResolveInfo,
        data: InfrahubComputedAttributeRecomputeInput,
        context: ContextInput | None = None,
    ) -> RecomputeComputedAttribute:
        graphql_context: GraphqlContext = info.context
        node_schema = registry.schema.get_node_schema(
            name=str(data.kind), branch=graphql_context.branch.name, duplicate=False
        )

        graphql_context.active_permissions.raise_for_permission(
            permission=ObjectPermission(
                namespace=node_schema.namespace,
                name=node_schema.name,
                action=PermissionAction.UPDATE.value,
                decision=PermissionDecision.ALLOW_DEFAULT.value
                if graphql_context.branch.name == registry.default_branch
                else PermissionDecision.ALLOW_OTHER.value,
            )
        )
        await apply_external_context(graphql_context=graphql_context, context_input=context)

        attribute = node_schema.get_attribute(name=str(data.attribute))

        if not attribute:
            raise ProcessingError(
                message=f"The indicated node does not have the specified attribute '{data.attribute}'"
            )
        if not attribute.computed_attribute:
            raise ProcessingError(
                message=f"The indicated node does not use a computed attribute for the specified attribute '{data.attribute}'"
            )

        recalculate_single_workflow = COMPUTED_ATTRIBUTE_PROCESS_JINJA2
        recalculate_all_workflow = TRIGGER_UPDATE_JINJA_COMPUTED_ATTRIBUTES
        if attribute.computed_attribute.kind == ComputedAttributeKind.TRANSFORM_PYTHON:
            if not await NodeManager.query(
                db=graphql_context.db,
                branch=graphql_context.branch,
                schema=CoreTransformPython,
                filters={"name__value": attribute.computed_attribute.transform},
            ):
                raise ProcessingError(
                    message=f"The transform for the indicated node computed attribute for the specified attribute '{data.attribute}' does not exist"
                )

            recalculate_single_workflow = COMPUTED_ATTRIBUTE_PROCESS_TRANSFORM
            recalculate_all_workflow = TRIGGER_UPDATE_PYTHON_COMPUTED_ATTRIBUTES

        if data.node_ids:
            nodes = await NodeManager.get_many(
                db=graphql_context.db, branch=graphql_context.branch, ids=list(data.node_ids)
            )
            for node in nodes.values():
                await get_workflow().submit_workflow(
                    workflow=recalculate_single_workflow,
                    context=graphql_context.get_context(),
                    parameters={
                        "branch_name": graphql_context.branch.name,
                        "computed_attribute_name": str(data.attribute),
                        "computed_attribute_kind": node_schema.kind,
                        "node_kind": node_schema.kind,
                        "object_id": node.id,
                        "context": context,
                    },
                )
        else:
            await get_workflow().submit_workflow(
                workflow=recalculate_all_workflow,
                context=graphql_context.get_context(),
                parameters={
                    "branch_name": graphql_context.branch.name,
                    "computed_attribute_name": str(data.attribute),
                    "computed_attribute_kind": node_schema.kind,
                    "context": context,
                },
            )

        result: dict[str, Any] = {"ok": True}
        return cls(**result)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Self

from graphene import Boolean, Enum, Field, InputObjectType, List, Mutation, String
from graphql import GraphQLResolveInfo

from infrahub import lock
from infrahub.core.account import GlobalPermission
from infrahub.core.branch import Branch
from infrahub.core.branch.enums import BranchStatus
from infrahub.core.constants import (
    CheckType,
    GlobalPermissions,
    InfrahubKind,
    PermissionDecision,
)
from infrahub.core.manager import NodeManager
from infrahub.core.protocols import CoreProposedChange
from infrahub.core.schema import NodeSchema
from infrahub.database import InfrahubDatabase, retry_db_transaction
from infrahub.events import (
    EventMeta,
    ProposedChangeApprovalRevokedEvent,
    ProposedChangeApprovedEvent,
    ProposedChangeRejectedEvent,
    ProposedChangeRejectionRevokedEvent,
)
from infrahub.exceptions import BranchNotFoundError, PermissionDeniedError, ValidationError
from infrahub.graphql.mutations.main import InfrahubMutationMixin
from infrahub.graphql.types.enums import CheckType as GraphQLCheckType
from infrahub.graphql.types.task import TaskInfo
from infrahub.lock import InfrahubLock
from infrahub.proposed_change.approval_revoker import do_revoke_approvals_on_updated_pcs
from infrahub.proposed_change.constants import ProposedChangeApprovalDecision, ProposedChangeState
from infrahub.proposed_change.models import RequestProposedChangePipeline
from infrahub.workers.dependencies import get_event_service
from infrahub.workflows.catalogue import PROPOSED_CHANGE_MERGE, REQUEST_PROPOSED_CHANGE_PIPELINE

from ...core.node.lock_utils import build_object_lock_name
from .main import InfrahubMutationOptions

if TYPE_CHECKING:
    from graphql import GraphQLResolveInfo

    from infrahub.core.node import Node
    from infrahub.events.models import InfrahubEvent

    from ..initialization import GraphqlContext

ProposedChangeApprovalDecisionInput = Enum.from_enum(ProposedChangeApprovalDecision)


class InfrahubProposedChangeMutation(InfrahubMutationMixin, Mutation):
    @classmethod
    def __init_subclass_with_meta__(
        cls, schema: NodeSchema, _meta: InfrahubMutationOptions | None = None, **options: dict[str, Any]
    ) -> None:
        # Make sure schema is a valid NodeSchema Node Class
        if not isinstance(schema, NodeSchema):
            raise ValueError(f"You need to pass a valid NodeSchema in '{cls.__name__}.Meta', received '{schema}'")

        if not _meta:
            _meta = InfrahubMutationOptions(cls)
        _meta.schema = schema

        super().__init_subclass_with_meta__(_meta=_meta, **options)

    @classmethod
    @retry_db_transaction(name="proposed_change_create")
    async def mutate_create(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        database: InfrahubDatabase | None = None,  # noqa: ARG003
        override_data: dict[str, Any] | None = None,
    ) -> tuple[Node, Self]:
        graphql_context: GraphqlContext = info.context

        override_data = {"created_by": {"id": graphql_context.active_account_session.account_id}}
        state = data.get("state", {}).get("value")
        if state and state != ProposedChangeState.OPEN.value:
            raise ValidationError(input_value="A proposed change has to be in the open state during creation")

        async with graphql_context.db.start_transaction() as dbt:
            proposed_change, result = await super().mutate_create(
                info=info, data=data, branch=branch, database=dbt, override_data=override_data
            )
            destination_branch = proposed_change.destination_branch.value
            source_branch = await _get_source_branch(db=dbt, name=proposed_change.source_branch.value)
            if destination_branch == source_branch.name:
                raise ValidationError(input_value="The source and destination branch can't be the same")
            if destination_branch != "main":
                raise ValidationError(
                    input_value="Currently only the 'main' branch is supported as a destination for a proposed change"
                )

        if graphql_context.service:
            request_proposed_change_model = RequestProposedChangePipeline(
                proposed_change=proposed_change.id,
                source_branch=source_branch.name,
                source_branch_sync_with_git=source_branch.sync_with_git,
                destination_branch=destination_branch,
            )

            await graphql_context.service.workflow.submit_workflow(
                workflow=REQUEST_PROPOSED_CHANGE_PIPELINE,
                parameters={"model": request_proposed_change_model},
                context=graphql_context.get_context(),
            )

        return proposed_change, result

    @classmethod
    async def mutate_update(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        database: InfrahubDatabase | None = None,  # noqa: ARG003
        node: Node | None = None,  # noqa: ARG003
    ) -> tuple[Node, Self]:
        graphql_context: GraphqlContext = info.context

        obj = await NodeManager.get_one_by_id_or_default_filter(
            db=graphql_context.db,
            kind=cls._meta.schema.kind,
            id=data.get("id"),
            branch=branch,
            include_owner=True,
            include_source=True,
        )
        state = ProposedChangeState(obj.state.value.value)
        state.validate_updatable()

        updated_state = None
        if state_update := data.get("state", {}).get("value"):
            updated_state = ProposedChangeState(state_update)
            state.validate_state_transition(updated_state)

        # Check if the draft state will change (defaults to current draft state)
        will_be_draft = data.get("is_draft", {}).get("value", obj.is_draft.value)

        # Check before starting a transaction, stopping in the middle of the transaction seems to break with memgraph
        if updated_state == ProposedChangeState.MERGED and graphql_context.account_session:
            try:
                graphql_context.active_permissions.raise_for_permission(
                    permission=GlobalPermission(
                        action=GlobalPermissions.MERGE_PROPOSED_CHANGE.value,
                        decision=PermissionDecision.ALLOW_ALL.value,
                    )
                )
            except PermissionDeniedError as exc:
                raise ValidationError(str(exc)) from exc

        if updated_state == ProposedChangeState.MERGED:
            if will_be_draft:
                raise ValidationError("A draft proposed change is not allowed to be merged")

            source_branch = await Branch.get_by_name(db=graphql_context.db, name=obj.source_branch.value)
            if source_branch.status == BranchStatus.NEED_UPGRADE_REBASE:
                raise ValidationError("The branch must be upgraded and rebased prior to merging the proposed change")

            data["state"]["value"] = ProposedChangeState.MERGING.value

        proposed_change, result = await super().mutate_update(
            info=info, data=data, branch=branch, database=graphql_context.db, node=obj
        )

        if updated_state == ProposedChangeState.MERGED:
            await graphql_context.service.workflow.execute_workflow(
                workflow=PROPOSED_CHANGE_MERGE,
                context=graphql_context.get_context(),
                parameters={
                    "proposed_change_id": proposed_change.id,
                    "proposed_change_name": proposed_change.name.value,
                },
            )
            # When the PROPOSED_CHANGE_MERGE succeeds it will have correctly changed the state
            # from the overridden "merging" value, so here we change it back to reflect the
            # correct value for the event that will be generated.
            proposed_change.node_changelog.attributes["state"].value = ProposedChangeState.MERGED.value

        return proposed_change, result


class ProposedChangeRequestRunCheckInput(InputObjectType):
    id = String(required=True)
    check_type = GraphQLCheckType(required=False)


class ProposedChangeRequestRunCheck(Mutation):
    class Arguments:
        data = ProposedChangeRequestRunCheckInput(required=True)

    ok = Boolean()

    @classmethod
    async def mutate(
        cls,
        root: dict,  # noqa: ARG003
        info: GraphQLResolveInfo,
        data: dict[str, Any],
    ) -> dict[str, bool]:
        graphql_context: GraphqlContext = info.context

        check_type = data.get("check_type") or CheckType.ALL

        identifier = data.get("id", "")
        proposed_change = await NodeManager.get_one_by_id_or_default_filter(
            id=identifier, kind=InfrahubKind.PROPOSEDCHANGE, db=graphql_context.db
        )
        state = ProposedChangeState(proposed_change.state.value.value)
        state.validate_state_check_run()

        destination_branch = proposed_change.destination_branch.value
        source_branch = await _get_source_branch(db=graphql_context.db, name=proposed_change.source_branch.value)

        request_proposed_change_model = RequestProposedChangePipeline(
            proposed_change=proposed_change.id,
            source_branch=source_branch.name,
            source_branch_sync_with_git=source_branch.sync_with_git,
            destination_branch=destination_branch,
            check_type=check_type,
        )
        if graphql_context.service:
            await graphql_context.service.workflow.submit_workflow(
                workflow=REQUEST_PROPOSED_CHANGE_PIPELINE,
                parameters={"model": request_proposed_change_model},
                context=graphql_context.get_context(),
            )

        return {"ok": True}


class ProposedChangeReviewInput(InputObjectType):
    id = String(required=True, description="The ID of the proposed change to review.")
    decision = ProposedChangeApprovalDecisionInput(
        required=True, description="The decision for the proposed change review."
    )


class ProposedChangeReview(Mutation):
    class Arguments:
        data = ProposedChangeReviewInput(required=True)

    ok = Boolean()

    @classmethod
    async def mutate(
        cls,
        root: dict,  # noqa: ARG003
        info: GraphQLResolveInfo,
        data: ProposedChangeReviewInput,
    ) -> dict[str, bool]:
        """
        This mutation is used to approve or reject a proposed change.
        It can also be used to undo an approval or rejection.
        """

        graphql_context: GraphqlContext = info.context
        graphql_context.active_permissions.raise_for_permission(
            permission=GlobalPermission(
                action=GlobalPermissions.REVIEW_PROPOSED_CHANGE.value, decision=PermissionDecision.ALLOW_ALL.value
            )
        )
        pc_id = str(data.id)
        lock_name = build_object_lock_name(pc_id)
        async with InfrahubLock(name=lock_name, connection=lock.registry.connection):
            proposed_change = await NodeManager.get_one_by_id_or_default_filter(
                id=pc_id, kind=CoreProposedChange, db=graphql_context.db, prefetch_relationships=True
            )
            state = ProposedChangeState(proposed_change.state.value.value)
            state.validate_reviewable()

            created_by = await proposed_change.created_by.get_peer(db=graphql_context.db)
            if created_by and created_by.id == graphql_context.active_account_session.account_id:
                raise ValidationError(input_value="You cannot review your own proposed changes")

            current_user = await NodeManager.get_one_by_id_or_default_filter(
                id=graphql_context.active_account_session.account_id,
                kind=InfrahubKind.GENERICACCOUNT,
                db=graphql_context.db,
            )

            async with graphql_context.db.start_session() as db:
                event = await cls._handle_decision(
                    db=db,
                    decision=data.decision,
                    proposed_change=proposed_change,
                    current_user=current_user,
                    context=graphql_context,
                )
                await proposed_change.save(db=db)

                if event:
                    event_service = await get_event_service()
                    await event_service.send(event=event)

        return {"ok": True}

    @classmethod
    async def _handle_decision(
        cls,
        db: InfrahubDatabase,
        decision: ProposedChangeApprovalDecision,
        proposed_change: CoreProposedChange,
        current_user: Node,
        context: GraphqlContext,
    ) -> InfrahubEvent | None:
        """Modify approved_by and rejected_by relationships of the prpoposed change based on the decision."""

        approved_by = await proposed_change.approved_by.get_peers(db=db)
        rejected_by = await proposed_change.rejected_by.get_peers(db=db)
        approved_by_ids = [node.id for _, node in approved_by.items()]
        rejected_by_ids = [node.id for _, node in rejected_by.items()]
        event: InfrahubEvent | None = None
        event_meta = EventMeta.from_context(context=context.get_context())

        match decision:
            case ProposedChangeApprovalDecision.APPROVE:
                if current_user.id in approved_by_ids:
                    raise ValidationError(input_value="You have already approved this proposed change")
                await proposed_change.approved_by.add(db=db, data=current_user)
                if current_user.id in rejected_by_ids:
                    await proposed_change.rejected_by.remove_locally(db=db, peer_id=current_user.id)

                event = ProposedChangeApprovedEvent(
                    proposed_change_id=proposed_change.id,
                    proposed_change_name=proposed_change.name.value,
                    proposed_change_state=proposed_change.state.value,
                    reviewer_account_id=current_user.id,
                    reviewer_account_name=current_user.name.value,
                    reviewer_decision=decision.value,
                    meta=event_meta,
                )

            case ProposedChangeApprovalDecision.CANCEL_APPROVE:
                if current_user.id not in approved_by_ids:
                    raise ValidationError(
                        input_value="You did not approve this proposed change yet, it can't be un-approved"
                    )
                await proposed_change.approved_by.remove_locally(db=db, peer_id=current_user.id)

                event = ProposedChangeApprovalRevokedEvent(
                    proposed_change_id=proposed_change.id,
                    proposed_change_name=proposed_change.name.value,
                    proposed_change_state=proposed_change.state.value,
                    reviewer_account_id=current_user.id,
                    reviewer_account_name=current_user.name.value,
                    reviewer_former_decision=ProposedChangeApprovalDecision.APPROVE.value,
                    meta=event_meta,
                )

            case ProposedChangeApprovalDecision.REJECT:
                if current_user.id in rejected_by_ids:
                    raise ValidationError(input_value="You have already rejected this proposed change")
                await proposed_change.rejected_by.add(db=db, data=current_user)
                if current_user.id in approved_by_ids:
                    await proposed_change.approved_by.remove_locally(db=db, peer_id=current_user.id)

                event = ProposedChangeRejectedEvent(
                    proposed_change_id=proposed_change.id,
                    proposed_change_name=proposed_change.name.value,
                    proposed_change_state=proposed_change.state.value,
                    reviewer_account_id=current_user.id,
                    reviewer_account_name=current_user.name.value,
                    reviewer_decision=decision.value,
                    meta=event_meta,
                )

            case ProposedChangeApprovalDecision.CANCEL_REJECT:
                if current_user.id not in rejected_by_ids:
                    raise ValidationError(
                        input_value="You did not reject this proposed change yet, it can't be un-rejected"
                    )
                await proposed_change.rejected_by.remove_locally(db=db, peer_id=current_user.id)

                event = ProposedChangeRejectionRevokedEvent(
                    proposed_change_id=proposed_change.id,
                    proposed_change_name=proposed_change.name.value,
                    proposed_change_state=proposed_change.state.value,
                    reviewer_account_id=current_user.id,
                    reviewer_account_name=current_user.name.value,
                    reviewer_former_decision=ProposedChangeApprovalDecision.REJECT.value,
                    meta=event_meta,
                )

            case _:
                raise ValidationError(input_value=f"Invalid decision {decision}")

        return event


class ProposedChangeMergeInput(InputObjectType):
    id = String(required=True)


class ProposedChangeMerge(Mutation):
    class Arguments:
        data = ProposedChangeMergeInput(required=True)
        wait_until_completion = Boolean(required=False)

    ok = Boolean()
    task = Field(TaskInfo, required=False)

    @classmethod
    async def mutate(
        cls,
        root: dict,  # noqa: ARG003
        info: GraphQLResolveInfo,
        data: dict[str, Any],
        wait_until_completion: bool = True,
    ) -> dict[str, bool]:
        graphql_context: GraphqlContext = info.context
        task: dict | None = None

        identifier = data.get("id", "")
        proposed_change = await NodeManager.get_one(
            id=identifier, kind=InfrahubKind.PROPOSEDCHANGE, db=graphql_context.db, raise_on_error=True
        )
        state = ProposedChangeState(proposed_change.state.value.value)
        if state != ProposedChangeState.OPEN:
            raise ValidationError("Only proposed change in OPEN state can be merged")

        async with graphql_context.db.start_session() as db:
            proposed_change.state.value = ProposedChangeState.MERGING.value
            await proposed_change.save(db=db)

        if wait_until_completion:
            await graphql_context.service.workflow.execute_workflow(
                workflow=PROPOSED_CHANGE_MERGE,
                context=graphql_context.get_context(),
                parameters={
                    "proposed_change_id": proposed_change.id,
                    "proposed_change_name": proposed_change.name.value,
                },
            )
        else:
            workflow = await graphql_context.service.workflow.submit_workflow(
                workflow=PROPOSED_CHANGE_MERGE,
                context=graphql_context.get_context(),
                parameters={
                    "proposed_change_id": proposed_change.id,
                    "proposed_change_name": proposed_change.name.value,
                },
            )
            task = {"id": workflow.id}

        return cls(ok=True, task=task)


class ProposedChangeCheckForApprovalRevokeInput(InputObjectType):
    ids = Field(List(of_type=String, required=True), required=False)


class ProposedChangeCheckForApprovalRevoke(Mutation):
    class Arguments:
        data = ProposedChangeCheckForApprovalRevokeInput(required=True)

    ok = Boolean()

    @classmethod
    async def mutate(
        cls,
        root: dict,  # noqa: ARG003
        info: GraphQLResolveInfo,
        data: dict[str, Any],
    ) -> dict[str, bool]:
        db = info.context.db
        ids: list[str] | None
        try:
            ids = data["ids"]
        except KeyError:
            ids = None

        await do_revoke_approvals_on_updated_pcs(db=db, proposed_changes_ids=ids)
        return cls(ok=True)


async def _get_source_branch(db: InfrahubDatabase, name: str) -> Branch:
    try:
        return await Branch.get_by_name(name=name, db=db)
    except BranchNotFoundError:
        raise ValidationError(
            input_value="The specified source branch for this proposed change was not found."
        ) from None
